# Copyright (C) 2019 SuperDARN Canada, University of Saskwatchewan
# Author: Marina Schmidt
# This code is improvement based on rti.py in the DaVitpy library
# https://github.com/vtsuperdarn/davitpy/blob/master/davitpy
#
# Modifications:
# 2021-05-12 Francis Tholley added gate2grounscatter to range-time plots
# 2021-06-18 Marina Schmidt (SuperDARN Canada) fixed ground scatter colour bug
# 2021-07-06 Carley Martin added keyword to aid in rounding start times
# 2022-03-04 Marina Schmidt added RangeEstimations in
# 2022-08-04 Carley Martin added elifs for HALF_SLANT options
# 2023-06-12 Carley Martin added coordinate plotting method
# 2023-06-28 Carley Martin refactored return values
# 2023-10-14 Carley Martin added embargoed data method
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#


"""
Range-Time Parameter (aka Intensity) plots
"""
import copy
import matplotlib.pyplot as plt
import numpy as np
import warnings

from datetime import datetime, timedelta
from matplotlib import dates, colors, colormaps, ticker
from typing import List

from pydarn import (RangeEstimation, check_data_type, Coords,
                    time2datetime, rtp_exceptions, plot_exceptions,
                    SuperDARNCpids, SuperDARNRadars, RadarID,
                    standard_warning_format, PyDARNColormaps,
                    determine_embargo, add_embargo)

warnings.formatwarning = standard_warning_format


class RTP:
    """
    Range-Time Parameter plots SuperDARN data using the following fields:

    Class pattern design: Builder
    This class inherits matplotlib.pyplot to inherit plotting features as well
    build off their builder design pattern.

    Notes
    -----
    This class inherits from matplotlib to generate the figures

    Methods
    -------
    plot_range_time
    plot_time_series
    plot_summary
    plot_coord_time
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_range_time()\n"\
                "   - plot_time_series()\n"\
                "   - plot_summary()\n"\
                "   - plot_coord_time()\n"

    @classmethod
    def plot_range_time(cls, dmap_data: List[dict], parameter: str = 'v',
                        beam_num: int = 0, channel: int = 'all', ax=None,
                        background: str = 'w', background_alpha: float = 1.0,
                        groundscatter: bool = False,
                        zmin: int = None, zmax: int = None,
                        start_time: datetime = None, end_time: datetime = None,
                        colorbar: plt.colorbar = None, ymin: int = None,
                        ymax: int = None, yspacing: int = 200,
                        range_estimation: RangeEstimation =
                        RangeEstimation.SLANT_RANGE,
                        colorbar_label: str = '',
                        norm=colors.Normalize, cmap: str = None,
                        filter_settings: dict = {},
                        date_fmt: str = '%y/%m/%d\n %H:%M',
                        round_start: bool = True, **kwargs):
        """
        Plots a range-time parameter plot of the given
        field name in the dmap_data

        Future Work
        -----------
        Support for other data input, like "time"
        dictionary key containing the datetime. However,
        further discussion is needed if this will be the keys
        name or maybe another input.

        Parameters
        -----------
        dmap_data: List[dict]
        parameter: str
            key name indicating which parameter to plot.
            Default: v (Velocity)
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The channel 0, 1, 2, 'all'
            Default : 'all'
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        groundscatter : boolean or str
            Flag to indicate if groundscatter should be plotted. If string
            groundscatter will be represented by that color else grey.
            Default : False
        background : str
            color of the background in the plot
            default: white
        background_alpha : float
            alpha (transparency) of the background in the plot
            default: 1.0 (opaque)
        zmin: int
            Minimum normalized value
            Default: minimum parameter value in the data set
        zmax: int
            Maximum normalized value
            Default: maximum parameter value in the data set
        ymax: int
            Sets the maximum y value
            Default: None, uses 'nrang' from data
        ymin: int
            sets the minimum y value
            Default: None, uses 0 range gate or minimum slant-range value
        yspacing: int
            sets the spacing between ticks
            Default: 200
        range_estimation: RangeEstimation
            set the y-axis to a desired range estimation calculation
            Default: RangeEstimation.SLANT_RANGE
        norm: matplotlib.colors.Normalization object
            This object use dependency injection to use any normalization
            method with the zmin and zmax.
            Default: colors.Normalization()
        start_time: datetime
            Start time of the plot x-axis as a datetime object
            Default: rounded to nearest hour-30 minutes from the first
                     record containing the chosen parameters data
        end_time: datetime
            End time of the plot x-axis as a datetime object
            Default: last record of the chosen parameters data
        date_fmt : str
            format of x-axis date ticks, follow datetime format
            Default: '%y/%m/%d\n %H:%M' (Year/Month/Day Hour:Minute)
        colorbar: matplotlib.pyplot.colorbar
            Setting a predefined colorbar for the range-time plot
            If None, then a colorbar will be created for the plot
            Default: None
        colorbar_label: str
            the label that appears next to the color bar
            Default: ''
        cmap: matplotlib.colormaps or str
            matplotlib colour map
            https://matplotlib.org/tutorials/colors/colormaps.html
            Default: PyDARNColormaps.PYDARN_VELOCITY
            note: to reverse the color just add _r to the string name
        filter_settings: dict
            dictionary of the following keys for filtering data out:
            max_array_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out any data points
                that is above this value.
            min_array_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out any data points that is
                below this value.
            max_scalar_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out data sections that is
                above this value.
            min_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is below this value.
            equal_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is does not equal the value.
        round_start: bool=True
                option to round the start time to give tick at start of xaxis
                Set True to round, set False to plot from start of data.
                Default: True
        kwargs:
            used for other methods in pyDARN
                - reflection_height
        Raises
        ------
        UnknownParameterError
        IncorrectPlotMethodError
        NoDataFoundError
        IndexError

        Returns
        -------
        im: matplotlib.pyplot.pcolormesh
            matplotlib object from pcolormesh
        cb: matplotlib.colorbar
            matplotlib color bar
        cmap: matplotlib.colormaps
            matplotlib color map object
        time_axis: list
            list representing the x-axis datetime objects
        y_axis: list
            list representing the y-axis range gates
        z_data: 2D numpy array
            2D array of the parameters values at the given time and range gate

        See Also
        ---------
        colors: https://matplotlib.org/2.0.2/api/colors_api.html
        color maps: PyDARNColormaps or
                    https://matplotlib.org/tutorials/colors/colormaps.html
        normalize:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.colors.Normalize.html
        colorbar:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.colorbar.html
        pcolormesh:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.pcolormesh.html

        """
        # Settings
        plot_filter = {'min_array_filter': dict(),
                       'max_array_filter': dict(),
                       'min_scalar_filter': dict(),
                       'max_scalar_filter': dict(),
                       'equal_scalar_filter': dict()}

        plot_filter.update(filter_settings)

        # If an axes object is not passed in then store
        # the equivalent object in matplotlib. This allows
        # for variant matplotlib plotting styles.
        if not ax:
            ax = plt.gca()

        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            index_first_match = next(i for i, d in enumerate(dmap_data)
                                     if parameter in d)
        except StopIteration:
            raise plot_exceptions.UnknownParameterError(parameter)
        cls.dmap_data = dmap_data
        check_data_type(cls.dmap_data, parameter, 'array', index_first_match)
        start_time, end_time = cls.__determine_start_end_time(start_time,
                                                              end_time)

        # y-axis coordinates, i.e., range gates,
        # TODO: implement variant other coordinate systems for the y-axis
        # y shape needs to be +1 longer as requirement of how pcolormesh
        # draws the pixels on the grid

        # because nrang can change based on mode we need to look
        # for the largest value
        y_max = max(record['nrang'] for record in cls.dmap_data)
        y = np.arange(0, y_max+1, 1)

        # z: parameter data mapped into the color mesh
        z = np.zeros((1, y_max)) * np.nan

        # x: time date data
        x = []

        # We cannot simply use numpy's built in min and max function
        # because of the groundscatter value :(

        # These flags indicate if zmin and zmax should change
        set_zmin = True
        set_zmax = True
        if zmin is None:
            zmin = cls.dmap_data[index_first_match][parameter][0]
            set_zmin = False
        if zmax is None:
            zmax = cls.dmap_data[index_first_match][parameter][0]
            set_zmax = False

        for dmap_record in cls.dmap_data:
            # get time difference to test if there is some gap data
            rec_time = time2datetime(dmap_record)
            diff_time = 0.0
            if rec_time > end_time:
                break
            if x != []:
                # 60.0 seconds in a minute
                delta_diff_time = abs(rec_time - x[-1])
                diff_time = delta_diff_time.seconds/60.0
                # Abs added above as some files have data out of order
                # abs stops the code from hanging and plotting over a day
                # of white space, but user needs to be warned that the
                # output may be incorrect
                if (rec_time - x[-1]) < timedelta(0):
                    # warnings are turned off for summary plots so
                    # for now print to console
                    # May be repeated, but will show what records are out
                    # of time order by doing so to help user
                    warnings.warn("Please be aware that the data for"
                                  " timestamp {} contains a record that is not"
                                  " in time order. As such the plot of the"
                                  " data may not be correct, you can solve"
                                  " this by sorting the data stream by date"
                                  " before plotting.".format(rec_time))

            # separation roughly 2 minutes
            if diff_time > 2.0:
                # if there is gap data (no data recorded past 2 minutes)
                # then fill it in with white space
                for _ in range(0, int(np.floor(diff_time/2.0))):
                    x.append(x[-1] + timedelta(0, 120))
                    i = len(x) - 1  # offset since we start at 0 not 1
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros([1, y_max]) * np.nan,
                                      axis=0)
            # Get data for the provided beam number
            if (beam_num == 'all' or dmap_record['bmnum'] == beam_num) and\
               (channel == 'all' or
                    dmap_record['channel'] == channel):
                if start_time <= rec_time:
                    # construct the x-axis array
                    # Numpy datetime is used because it properly formats on the
                    # x-axis
                    x.append(rec_time)
                    # I do this to avoid having an extra loop to just count how
                    # many records contain the beam number
                    i = len(x) - 1  # offset since we start at 0 not 1

                    # insert a new column into the z_data
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros([1, y_max]) * np.nan,
                                      axis=0)
                    try:
                        if len(dmap_record[parameter]) == dmap_record['nrang']:
                            good_gates = range(len(dmap_record[parameter]))
                        else:
                            good_gates = dmap_record['slist']

                        # get the range gates that have "good" data in it
                        for j in range(len(good_gates)):
                            # if it is groundscatter store a very
                            # low number in that cell
                            if groundscatter and\
                               dmap_record['gflg'][j] == 1:
                                # chosen value from davitpy to make the
                                # groundscatter a different color
                                # from the color map
                                z[i][good_gates[j]] = -1000000
                            # otherwise store parameter value
                            # TODO: refactor and clean up this code
                            elif cls.__filter_data_check(dmap_record,
                                                         plot_filter, j):
                                z[i][good_gates[j]] = \
                                        dmap_record[parameter][j]
                                # calculate min and max value
                                if not set_zmin and\
                                   z[i][good_gates[j]] < zmin:
                                    zmin = z[i][good_gates[j]]
                                if not set_zmax and \
                                   z[i][good_gates[j]] > zmax:
                                    zmax = z[i][good_gates[j]]
                    # a KeyError may be thrown because slist is not created
                    # due to bad quality data.
                    except KeyError:
                        continue
        x.append(end_time)
        # Check if there is any data to plot
        if np.all(np.isnan(z)):
            raise plot_exceptions.\
                    NoDataFoundError(parameter, beam_num,
                                     start_time=start_time,
                                     end_time=end_time,
                                     opt_beam_num=cls.dmap_data[0]['bmnum'])
        if range_estimation != RangeEstimation.RANGE_GATE:
            # Get rxrise from hardware files (consistent with RST)
            rxrise = SuperDARNRadars.radars[RadarID(cls.dmap_data[0]['stid'])]\
                                    .hardware_info.rx_rise_time
            frang = int(cls.dmap_data[0]['frang'])
            rsep = int(cls.dmap_data[0]['rsep'])

            y = range_estimation(frang=frang, rxrise=rxrise,
                                 rsep=rsep, nrang=y_max, **kwargs)

            y0inx = np.min(np.where(np.isfinite(y))[0])
            y = y[y0inx:]
            z = z[:, y0inx:]

        time_axis, y_axis = np.meshgrid(x, y)
        z_data = np.ma.masked_where(np.isnan(z.T), z.T)
        Default = {'noise.sky': (1e0, 1e5),
                   'tfreq': (8, 22),
                   'nave': (0, 60),
                   'p_l': (0, 45),
                   'v': (-200, 200),
                   'w_l': (0, 250),
                   'elv': (0, 45)}

        if np.isinf(zmin) and zmin < 0:
            zmin = Default[parameter][0]
            warnings.warn("Warning: zmin is -inf, set zmin to {}. You can"
                          "set zmin and zmax in the function"
                          " options".format(zmin))
        if np.isinf(zmax) and zmax > 0:
            zmax = Default[parameter][1]
            warnings.warn("Warning: zmax is inf, set zmax to {}. You can"
                          "set zmin and zmax in the functions"
                          " options".format(zmax))
        norm = norm(zmin, zmax)
        if isinstance(cmap, str):
            cmap = colormaps.get_cmap(cmap)
        else:
            cmaps = {'p_l': PyDARNColormaps.PYDARN_PLASMA,
                     'v': PyDARNColormaps.PYDARN_VELOCITY,
                     'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                     'elv': PyDARNColormaps.PYDARN_INFERNO}
            cmap = cmaps[parameter]

        # set the background color, this needs to happen to avoid
        # the overlapping problem that occurs
        cmap.set_bad(color=background, alpha=background_alpha)
        # plot!
        im = ax.pcolormesh(time_axis, y_axis, z_data, lw=0.01,
                           cmap=cmap, norm=norm, **kwargs)

        if isinstance(groundscatter, str):
            ground_scatter = np.ma.masked_where(z_data != -1000000, z_data)
            gs_color = colors.ListedColormap([groundscatter])
            ax.pcolormesh(time_axis, y_axis, ground_scatter, lw=0.01,
                          cmap=gs_color, norm=norm, **kwargs)

        elif groundscatter:
            ground_scatter = np.ma.masked_where(z_data != -1000000, z_data)
            gs_color = colors.ListedColormap(['grey'])
            ax.pcolormesh(time_axis, y_axis, ground_scatter, lw=0.01,
                          cmap=gs_color, norm=norm, **kwargs)

        # setup some standard axis information
        if ymax is None:
            ymax = np.max(y)

        if ymin is None:
            ymin = np.min(y)

        ax.set_ylim(ymin, ymax)

        if (range_estimation != RangeEstimation.RANGE_GATE and
            range_estimation != RangeEstimation.TIME_OF_FLIGHT):
            ax.yaxis.set_ticks(np.arange(np.ceil(ymin/100.0)*100,
                                         ymax+1, yspacing))
        elif range_estimation == RangeEstimation.RANGE_GATE:
            ax.yaxis.set_ticks(np.arange(ymin, ymax+1, (ymax)/5))
        else:
            ax.yaxis.set_ticks(np.arange(0, np.ceil(ymax)+1, 5))

        # SuperDARN file typically are in 2hr or 24 hr files
        # to make the minute ticks sensible, the time length is detected
        # then a interval is picked. 30 minute ticks for 24 hr plots
        # and 5 minute ticks for 2 hour plots.
        data_time_length = end_time - start_time
        # 3 hours * 60 minutes * 60 seconds
        if data_time_length.total_seconds() > 3*60*60:
            tick_interval = 30
        else:
            tick_interval = 1
        # byminute keyword makes sure that the ticks are situated at
        # the minute or half hour marks, rather than at a set interval
        ax.xaxis.set_minor_locator(
            dates.MinuteLocator(byminute=range(0, 60, tick_interval)))

        # Upon request of Daniel Billet and others, I am rounding
        # the time down so the plotting x-axis will show the origin
        # time label
        # Updated to give option to round down and make sure
        # rounding to same frequency as plot axis ticks if less than 1 hour
        if round_start:
            major_locator, _ = plt.xticks()
            dt = dates.num2date(major_locator[1]) -\
                dates.num2date(major_locator[0])
            tick_sep = dt.seconds//60
            if tick_sep > 0:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % tick_sep,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
            else:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % 15,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
        else:
            rounded_down_start_time = x[0]

        ax.set_xlim([rounded_down_start_time, x[-1]])
        ax.xaxis.set_major_formatter(dates.DateFormatter(date_fmt))

        if range_estimation != RangeEstimation.RANGE_GATE:
            ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(2))
        else:
            ax.yaxis.set_minor_locator(ticker.MultipleLocator(5))
        # so the plots gets to the ends
        ax.margins(0)

        # Create color bar if None supplied
        if not colorbar:
            with warnings.catch_warnings():
                warnings.filterwarnings('error')
                try:
                    if isinstance(norm, colors.LogNorm):
                        if zmin == 0:
                            cb = ax.figure.colorbar(im, ax=ax, extend='max')
                        else:
                            cb = ax.figure.colorbar(im, ax=ax, extend='both')
                    else:
                        locator = ticker.MaxNLocator(symmetric=True,
                                                     min_n_ticks=3,
                                                     integer=True,
                                                     nbins='auto')
                        ticks = locator.tick_values(vmin=zmin, vmax=zmax)
                        if zmin == 0:
                            cb = ax.figure.colorbar(im, ax=ax, extend='max',
                                                    ticks=ticks)
                        else:
                            cb = ax.figure.colorbar(im, ax=ax, extend='both',
                                                    ticks=ticks)

                except (ZeroDivisionError, Warning):
                    raise rtp_exceptions.RTPZeroError(parameter, beam_num,
                                                      zmin, zmax,
                                                      norm) from None
        else:
            cb = colorbar
        if colorbar_label != '':
            cb.set_label(colorbar_label)

        if determine_embargo(end_time, dmap_data[-1]['cp'],
                             SuperDARNRadars.radars[RadarID(dmap_data[-1]['stid'])].name):
            add_embargo(plt.gcf())

        return {'ax': ax,
                'ccrs': None,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'plot_data': im,
                         'x': x,
                         'y': y,
                         'z': z_data}
                }

    @classmethod
    def plot_time_series(cls, dmap_data: List[dict],
                         parameter: str = 'tfreq', beam_num: int = 0,
                         ax=None, gate: int = 0, start_time: datetime = None,
                         end_time: datetime = None,
                         date_fmt: str = '%y/%m/%d\n %H:%M',
                         channel='all', scale: str = 'linear',
                         cp_name: bool = True, color: str = 'black',
                         linestyle: str = '-', linewidth: float = 1,
                         round_start: bool = True, **kwargs):
        """
        Plots the time series of a scalar parameter

        Parameters
        ----------
        dmap_data : List[dict]
            List of dictionaries representing SuperDARN data
        parameter : str
            Scalar parameter to plot
            Default: tfreq
        beam_num : int
            The beam number of data to plot
            Default: 0
        ax : matplotlib axes object
            option to pass in axes object from matplotlib.pyplot
            Default: plt.gca()
        start_time: datetime
            Start time of the plot x-axis as a datetime object
            Default: first record date
        end_time: datetime
            End time of the plot x-axis as a datetime object
            Default: last record date
        date_fmt : datetime format string
            Date format for the x-axis
            Default: '%y/%m/%d \n %H:%M'
        channel : int or str
            integer indicating which channel to plot or 'all' to
            plot all channels
            Default: 'all'
        scale: str
            The y-axis scaling. This is not used for plotting the cp ID
            Default: log
        cp_name : bool
            If True, the cp ID name will be printed
            along side the number. Otherwise the cp ID will
            just be printed. This is only used for the parameter cp
            Default: True
        round_start: bool=True
                option to round the start time to give tick at start of xaxis
                Set True to round, set False to plot from start of data.
                Default: True
        color: str
            color of the line
            default: black
        linestyle: str
            style of line with dash marks or solid
            default: solid
        linewidth: float
            the width of the line
            default: 1
        kwargs
            kwargs passed into plot_date

        Raises
        ------
        UnknownParameterError
        IncorrectPlotMethodError
        NoDataFoundError
        IndexError

        Returns
        -------
        lines: list
            list of matplotlib.lines.Lines2D object representing the
            time-series data if plotting parameter cp then it will be None
        x: list
            list of datetime objects representing x-axis time series
        y: list
            list of scalar values for each datetime object

        See Also
        --------
        yscale:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.yscale.html
        plot_date:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.plot_date.html
        colors: https://matplotlib.org/2.0.2/api/colors_api.html
        """
        # check if axes object is passed in, if not
        # Default to plt.gca()
        if not ax:
            ax = plt.gca()

        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            next(i for i, d in enumerate(dmap_data) if parameter in d)
        except StopIteration:
            raise plot_exceptions.UnknownParameterError(parameter)

        cls.dmap_data = dmap_data
        start_time, end_time = cls.__determine_start_end_time(start_time,
                                                              end_time)

        # initialized here for return purposes
        lines = None
        # parameter data
        y = []
        # date time
        x = []
        # plot CPID
        if parameter == 'cp':
            old_cpid = None
            for dmap_record in cls.dmap_data:
                # TODO: this check could be a function call
                x.append(time2datetime(dmap_record))

                if (dmap_record['bmnum'] == beam_num or beam_num == 'all') and\
                   (dmap_record['channel'] == channel or channel == 'all'):
                    rec_time = time2datetime(dmap_record)
                    if start_time <= rec_time and rec_time <= end_time:
                        if old_cpid != dmap_record['cp'] or old_cpid is None:
                            ax.axvline(x=rec_time, color='black')
                            old_cpid = dmap_record['cp']
                            ax.text(x=rec_time + timedelta(seconds=600), y=0.7,
                                    s=dmap_record['cp'])
                            if cp_name:
                                # Keeping this commented code in to show how
                                # we could get the name from the file; however,
                                # there is not set format for combf field ...
                                # so we will use the dictionary to prevent
                                # errors or incorrect names on the plot.
                                # However, we should get it from the file
                                # not a dictionary that might not be updated
                                # cpid_command =
                                #   dmap_record['combf'].split(' ')
                                # if len(cpid_command) == 1:
                                #     cp_name = cpid_command[0]
                                # elif len(cpid_command) == 0:
                                #     cp_name = 'unknown'
                                # else:
                                #     cp_name = cpid_command[1]
                                if dmap_record['cp'] < 0:
                                    cpID_name = 'discretionary \n{}'\
                                            ''.format(SuperDARNCpids.cpids.
                                                      get(abs(
                                                          dmap_record['cp']),
                                                          'unknown'))
                                else:
                                    cpID_name =\
                                            SuperDARNCpids.cpids.\
                                            get(abs(dmap_record['cp']),
                                                'unknown')
                                ax.text(x=rec_time + timedelta(seconds=600),
                                        y=0.1, s=cpID_name)

            # Check if the old cp ID change, if not then there was no data
            if old_cpid is None:
                raise plot_exceptions.\
                        NoDataFoundError(parameter, beam_num,
                                         start_time=start_time,
                                         end_time=end_time,
                                         opt_beam_num=cls.
                                         dmap_data[0]['bmnum'])

            # to get rid of y-axis numbers
            ax.set_yticks([])
        else:
            for dmap_record in cls.dmap_data:
                # TODO: this check could be a function call
                rec_time = time2datetime(dmap_record)
                if start_time <= rec_time and rec_time <= end_time:
                    if (dmap_record['bmnum'] == beam_num or
                        beam_num == 'all') and \
                       (channel == dmap_record['channel'] or channel == 'all'):
                        # construct the x-axis array
                        x.append(rec_time)
                        try:
                            if parameter == 'tfreq':
                                # Convert kHz to MHz by dividing by 1000
                                y.append(dmap_record[parameter]/1000)
                            elif isinstance(dmap_record[parameter],
                                            np.ndarray):
                                if gate in dmap_record['slist']:
                                    for i in range(len(dmap_record['slist'])):
                                        if dmap_record['slist'][i] == gate:
                                            break
                                    y.append(dmap_record[parameter][i])
                                else:
                                    y.append(np.ma.masked)
                            else:
                                y.append(dmap_record[parameter])
                        except KeyError:
                            y.append(np.ma.masked)
                    # else plot missing data
                    elif len(x) > 0:
                        diff_time = rec_time - x[-1]
                        # if the time difference is greater than 2 minutes
                        # meaning no data was collected for that time period
                        # then plot nothing.
                        if diff_time.total_seconds() > 2.0 * 60.0:
                            x.append(rec_time)
                            y.append(np.nan)  # for masking the data
            # Check if there is any data to plot
            if np.all(np.isnan(y)) or len(x) == 0:
                raise plot_exceptions.\
                        NoDataFoundError(parameter, beam_num,
                                         start_time=start_time,
                                         end_time=end_time,
                                         opt_beam_num=cls.
                                         dmap_data[0]['bmnum'])

            # using masked arrays to create gaps in the plot
            # otherwise the lines will connect in gapped data
            my = np.ma.array(y)
            my = np.ma.masked_where(np.isnan(my), my)

            lines = ax.plot(x, my, color=color, linestyle=linestyle,
                            linewidth=linewidth)

            if round_start:
                major_locator, _ = plt.xticks()
                dt = dates.num2date(major_locator[1]) -\
                    dates.num2date(major_locator[0])
                tick_sep = dt.seconds//60
                if tick_sep > 0:
                    rounded_down_start_time = x[0] -\
                        timedelta(minutes=x[0].minute % tick_sep,
                                  seconds=x[0].second,
                                  microseconds=x[0].microsecond)
                else:
                    rounded_down_start_time = x[0] -\
                        timedelta(minutes=x[0].minute % 15,
                                  seconds=x[0].second,
                                  microseconds=x[0].microsecond)
            else:
                rounded_down_start_time = x[0]

            ax.set_xlim([rounded_down_start_time, x[-1]])
            ax.set_yscale(scale)

        # set date format and minor hourly locators
        # Rounded the time down to show origin label upon
        # Daniel Billet and others request.
        # TODO: may move this to its own function
        if round_start:
            major_locator, _ = plt.xticks()
            dt = dates.num2date(major_locator[1]) -\
                dates.num2date(major_locator[0])
            tick_sep = dt.seconds//60
            if tick_sep > 0:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % tick_sep,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
            else:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % 15,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
        else:
            rounded_down_start_time = x[0]

        ax.set_xlim([rounded_down_start_time, x[-1]])

        ax.xaxis.set_major_formatter(dates.DateFormatter(date_fmt))
        ax.xaxis.set_minor_locator(dates.HourLocator())
        # SuperDARN file typically are in 2hr or 24 hr files
        # to make the minute ticks sensible, the time length is detected
        # then a interval is picked. 30 minute ticks for 24 hr plots
        # and 5 minute ticks for 2 hour plots.
        data_time_length = end_time - start_time
        # 3 hours * 60 minutes * 60 seconds
        if data_time_length.total_seconds() > 3*60*60:
            tick_interval = 30
        else:
            tick_interval = 1
        # byminute keyword makes sure that the ticks are situated at
        # the minute or half hour marks, rather than at a set interval
        ax.xaxis.set_minor_locator(
            dates.MinuteLocator(byminute=range(0, 60, tick_interval)))

        ax.margins(x=0)
        ax.tick_params(axis='y', which='minor')

        if determine_embargo(end_time, dmap_data[-1]['cp'],
                             SuperDARNRadars.radars[RadarID(dmap_data[-1]['stid'])].name):
            add_embargo(plt.gcf())

        return {'ax': ax,
                'ccrs': None,
                'cm': None,
                'cb': None,
                'fig': plt.gcf(),
                'data': {'lines': lines,
                         'x': x,
                         'y': y}
                }

    @classmethod
    def plot_summary(cls, dmap_data: List[dict],
                     beam_num: int = 0, figsize: tuple = (11, 8.5),
                     watermark: bool = False, boundary: dict = {},
                     cmaps: dict = {}, lines: dict = {},
                     plot_elv: bool = True, title=None,
                     background: str = 'w',
                     groundscatter: bool = True,
                     channel: int = 'all', line_color: dict = {},
                     range_estimation: object =
                     RangeEstimation.SLANT_RANGE,
                     latlon: str = None, coords: object = Coords.AACGM,
                     vector_parameters: list = [('p_l'), ('v'),
                                                ('w_l'), ('elv')],
                     **kwargs):
        """
        Plots the summary of several SuperDARN parameters using time-series and
        range-time plots. Please see Notes for further description
        on what is plotted.

        Future Work
        ------------
        day-night terminators

        Parameters
        ----------
        dmap_data: List[dict]
            List of dictionaries of the data to be plotted containing the
            parameter fields used in the summary plot.
        beam_num : int
            beam number to plot
            default: 0
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        background: string
            background color of the plots
            default: white
        groundscatter : boolean
            Flag to indicate if groundscatter should be plotted.
            Placed only on the velocity plot.
            Default : True
        channel : int
            channel number that will be plotted
            in the summary plot.
            Default: 'all'
        line_color: Dict
            dictionary of time series parameters and the line color you
            require for each
            e.g.: {'nave': 'b', 'tfreq': 'r'}
            Default {} (set in code later)
        range_estimation: RangeEstimation object
            set the y-axis to a desired  range gate estimation
            calculation
            Default: RangeEstimation.SLANT_RANGE
        figsize : (int,int)
            tuple containing (height, width) figure size
            Default: 11 x 8.5
        watermark : boolean
            text that runs across the plot stating "Not for Publication Use"
            default: True
        cmaps: dict or str
            dictionary of matplotlib color maps for the summary
            range time parameter plots.
            https://matplotlib.org/tutorials/colors/colormaps.html
            Default: {'p_l': PyDARNColormaps.PYDARN_PLASMA,
                      'v': PyDARNColormaps.PYDARN_VELOCITY,
                      'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                      'elv': PyDARNColormaps.PYDARN_INFERNO}
            note: to reverse the color just add _r to the string name
        lines: dict or str
            dictionary of time-series line colors.
            Default: black
        boundary: dict
            tuple as the value (zmin, zmax) for the plots
            Default: {'noise.sky': (1e0, 1e5),
                      'tfreq': (8, 22),
                      'nave': (0, 60),
                      'p_l': (0, 45),
                      'v': (-200, 200),
                      'w_l': (0, 250),
                      'elv': (0, 45)}
        plot_elv: boolean
            boolean determines if elevation should be plotted or not.
            If there is no data for elevation data field then elevation is not
            plotted.
            Default: True
        title: str
            title of the plot
            Default: auto-generated by the files details
            {radar name} {Radar system (if applicable)} Fitacf {version}
            {start hour/date} - {end hour/date}  Beam {number}
        coords: Coords
            set default coordinate system for the coordinate time plot
            default: Coords.AACGM
        latlon: str
            set which coordinate you want to plot with
            default: None
        vector_parameters: list 
            Required parameters for plotting in the summary plot
            Must be a subset of the default list below
            default:[('p_l'), ('v'), ('w_l'), ('elv')]
        kwargs:
            reflection_height for ground_scatter_mapped method
            background
        Raises
        ------
        IndexError
        UnknownParameterError
        IncorrectPlotMethodError
        NoDataFoundError

        See Also
        --------
        plot_range_time : plots range-time parameter
        plot_time_series : plots time-series

        Return
        ------
        fig: matplotlib.pyplot.figure object
        axes: list
            list of matplotlib.pyplot.axes objects to generate
            the subplots in the summary plot

        Notes
        -----
        These are the following parameters in the SuperDARN FitACF file that is
        plotted in a summary plot:
            - noise.sky :  (time-series)
            - tfreq : transmission frequency (time-series)
            - nave : number of averages  (time-series)
            - cp : control program ID (time-series)
            - p_l : Signal to Noise ratio (range-time)
            - v : velocity (range-time)
            - w_l : spectral width (range-time)
            - elv : elevation (optional) (range-time)
        """

        # Default boundary ranges for the various parameter
        boundary_ranges = {'noise.search': (1e0, 1e5),
                           'noise.sky': (1e0, 1e5),
                           'tfreq': (8, 22),
                           'nave': (0, 60),
                           'p_l': (0, 45),
                           'v': (-200, 200),
                           'w_l': (0, 250),
                           'elv': (0, 45)}
        boundary_ranges.update(boundary)

        # Default color maps for the summary plot
        line = {'noise.search': '--',
                'noise.sky': '-',
                'tfreq': '--',
                'nave': '-'}
        color = {'noise.search': 'k',
                 'noise.sky': 'k',
                 'tfreq': 'k',
                 'nave': 'b'}

        if isinstance(line_color, dict):
            color.update(line_color)
        else:
            color.update({k: lines for k, v in line_color.items()})

        if isinstance(lines, dict):
            line.update(lines)
        else:
            line.update({k: lines for k, v in line.items()})
        cmap = {'p_l': PyDARNColormaps.PYDARN_PLASMA,
                'v': PyDARNColormaps.PYDARN_VELOCITY,
                'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                'elv': PyDARNColormaps.PYDARN_INFERNO}
        if isinstance(cmaps, dict):
            cmap.update(cmaps)
        else:
            cmap.update({k: cmaps for k, v in cmap.items()})

        fig = plt.figure(figsize=figsize)

        # axes objects in order of creation:
        # [noise, tfreq, cp, snr, vel, spect, elv]
        # Check if the radar has elevation information if not
        # do not plot elevation
        scalar_parameters = [('noise.sky', 'noise.search'), ('tfreq', 'nave'),
                             ('cp')]

        try:
            # need to use any because some records at the start
            # can be partial which doesn't mean there is no elv
            # data
            if any('elv' in d for d in dmap_data) and plot_elv\
                    and ('elv' in vector_parameters):
                num_plots = len(scalar_parameters) + len(vector_parameters)
            elif 'elv' not in vector_parameters:
                num_plots = len(scalar_parameters) + len(vector_parameters)
            else:
                num_plots = len(scalar_parameters) + len(vector_parameters) - 1
                vector_parameters.remove('elv')
        except KeyError:
            num_plots = len(scalar_parameters) + len(vector_parameters) - 1
            vector_parameters.remove('elv')
        axes = []

        # Check integrity of vector_parameters before committing to plotting
        if not isinstance(vector_parameters, list):
            raise Exception('The summary plot needs a formatted list of vector '
                            'parameters to plot. E.G. ["v", "p_l"]. '
                            'Summary plots allow plotting of the following '
                            'vector parameters: v (velocity), '
                            'p_l (Signal to noise ratio), '
                            'w_l (spectral width), '
                            'elv (elevation).')
        # Test for elevation data being removed
        if not vector_parameters or len(vector_parameters) == 0:
            raise Exception('The summary plot needs a list of vector '
                            'parameters to plot. The file you have chosen to '
                            'plot may not have elevation data. '
                            'Summary plots allow plotting of the following '
                            'vector parameters: v (velocity), '
                            'p_l (Signal to noise ratio), '
                            'w_l (spectral width), '
                            'elv (elevation).')
        # Test for non standard parameters
        allowed_parameters = ['elv', 'v', 'w_l', 'p_l']
        test_list = [i for i in vector_parameters if i not in allowed_parameters]
        if len(test_list) > 0 :
            raise Exception('Summary plots allow plotting of the following '
                            'vector parameters only: v (velocity), '
                            'p_l (Signal to noise ratio), '
                            'w_l (spectral width), '
                            'elv (elevation).')

        # List of parameters plotted in the summary plot, tuples are used
        # for shared plot parameters like noise.search and noise.sky
        axes_parameters = scalar_parameters + vector_parameters

        # labels to show on the summary plot for each parameter
        labels = {'noise.sky': 'Sky \n Noise', 
                  'noise.search': 'Search\n Noise',
                  'tfreq': 'Freq\n ($MHz$)',
                  'nave': 'Nave',
                  'cp': 'CP ID',
                  'p_l': 'SNR ($dB$)',
                  'v': 'Velocity\n ($m\ s^{-1}$)',
                  'w_l': 'Spectral Width\n ($m\ s^{-1}$)',
                  'elv': 'Elevation\n ($\degree$)'}


        # TODO: We need a cleaner solution to this and still allow user input
        # for figure size. For now it 'works'. Thanks to P. Pitzer for current
        # solution.
        start_pos = [0,0,0, 0.60, 0.75, 0.80, 0.84, 0.87, 0.90, 0.92]
        for i in range(num_plots):
            # time-series plots
            # position: [left, bottom, width, height]
            if i < 3:
                axes.append(fig.add_axes([0.1,
                                          (start_pos[num_plots]) -
                                          (i / num_plots * .5),
                                          0.76,
                                          0.06 * (7 / num_plots)]))
            # range-time plots
            else:
                axes.append(fig.add_axes([0.1,
                                          1.04 - (i / num_plots * 1.1),
                                          0.95,
                                          0.14 * (7 / num_plots)]))

        for i in range(num_plots):
            # plot time-series
            if i < 2:
                # for noise.search and frequency plots as they share x-axis
                # with noise.sky and nave
                if i == 0:
                    scale = 'log'
                else:
                    scale = 'linear'

                # plot time-series parameters that share a plot
                if i < 2:
                    with warnings.catch_warnings():
                        # Only show the first warning of each type so we don't
                        # get four of each warnings in summary plots
                        warnings.simplefilter("once")
                        cls.plot_time_series(dmap_data, beam_num=beam_num,
                                             parameter=axes_parameters[i][0],
                                             scale=scale, channel=channel,
                                             color=color[
                                                 axes_parameters[i][0]],
                                             ax=axes[i],
                                             linestyle=line[
                                                 axes_parameters[i][0]],
                                             label=labels[
                                                 axes_parameters[i][0]],
                                                 **kwargs)
                    axes[i].set_ylabel(labels[axes_parameters[i][0]],
                                       rotation=0, labelpad=30)
                    axes[i].\
                        axhline(y=boundary_ranges[axes_parameters[i][0]][0] +
                                0.8, xmin=-0.11, xmax=-0.05, clip_on=False,
                                color=color[axes_parameters[i][0]],
                                linestyle=line[axes_parameters[i][0]])
                    axes[i].set_ylim(boundary_ranges[axes_parameters[i][0]][0],
                                     boundary_ranges[axes_parameters[i][0]][1])
                    # For better y-axis ticks
                    if scale == 'log':
                        axes[i].yaxis.set_major_locator(ticker.
                                                        LogLocator(numticks=3))
                    else:
                        axes[i].yaxis.\
                                set_major_locator(ticker.
                                                  MaxNLocator(integer=True,
                                                              nbins=3))
                    axes[i].yaxis.set_label_coords(-0.08, 0.085)

                    if i == 1:
                        # plot the shared parameter
                        second_ax = axes[i].twinx()
                        with warnings.catch_warnings():
                            warnings.simplefilter("once")
                            cls.plot_time_series(dmap_data, beam_num=beam_num,
                                                 parameter=axes_parameters[
                                                     i][1],
                                                 color=color[axes_parameters[
                                                     i][1]],
                                                 channel=channel,
                                                 scale=scale, ax=second_ax,
                                                 linestyle=line[
                                                     axes_parameters[i][1]],
                                                 **kwargs)
                        second_ax.set_xticklabels([])
                        second_ax.set_ylabel(labels[axes_parameters[i][1]],
                                             rotation=0,
                                             labelpad=25, color=color[
                                                axes_parameters[i][1]])
                        second_ax.\
                            axhline(y=boundary_ranges[axes_parameters[i][1]][0]
                                    + 0.8, xmin=1.07, xmax=1.13,
                                    clip_on=False,
                                    linestyle=line[axes_parameters[i][1]],
                                    color=color[axes_parameters[i][1]])
                        second_ax.\
                            set_ylim(boundary_ranges[axes_parameters[i][1]][0],
                                     boundary_ranges[axes_parameters[i][1]][1])
                        second_ax.yaxis.set_label_coords(1.1, 0.7)
                        # Set color of second axis
                        second_ax.spines["right"].set_edgecolor(color=color[
                                                    axes_parameters[i][1]])
                        second_ax.tick_params(axis='y', color=color[
                                                axes_parameters[i][1]])
                        [lab.set_color(color=color[axes_parameters[i][1]])
                            for lab in second_ax.yaxis.get_ticklabels()]

                        if scale == 'log':
                            second_ax.yaxis.\
                                    set_major_locator(ticker.
                                                      LogLocator(numticks=4))

                        else:
                            second_ax.yaxis.\
                                    set_major_locator(ticker.
                                                      MaxNLocator(integer=True,
                                                                  nbins=3))

                axes[i].set_facecolor(background)
            # plot cp id
            elif i == 2:
                with warnings.catch_warnings():
                    warnings.simplefilter("once")
                    cls.plot_time_series(dmap_data, beam_num=beam_num,
                                         channel=channel,
                                         parameter=axes_parameters[i],
                                         ax=axes[i], **kwargs)
                axes[i].set_ylabel('CPID', rotation=0, labelpad=30)
                axes[i].yaxis.set_label_coords(-0.08, 0.079)
                axes[i].set_facecolor(background)
            # plot range-time
            else:
                # Current standard is to only have groundscatter
                # on the velocity plot.
                if groundscatter and axes_parameters[i] == 'v':
                    grndflg = True
                else:
                    grndflg = False
                # with warning catch, catches all the warnings
                # that would be produced by time-series this would be
                # the citing warning.
                with warnings.catch_warnings():
                    warnings.simplefilter("once")
                    if latlon is None:
                        rt_rtn =\
                            cls.plot_range_time(dmap_data, beam_num=beam_num,
                                            colorbar_label=labels[
                                                axes_parameters[i]],
                                            channel=channel,
                                            parameter=axes_parameters[i],
                                            ax=axes[i], groundscatter=grndflg,
                                            cmap=cmap[axes_parameters[i]],
                                            zmin=boundary_ranges[
                                                axes_parameters[i]][0],
                                            zmax=boundary_ranges[
                                                axes_parameters[i]][1],
                                            yspacing=500,
                                            background=background,
                                            range_estimation=range_estimation,
                                            **kwargs)
                    else:
                        rt_rtn =\
                            cls.plot_coord_time(dmap_data, beam_num=beam_num,
                                            colorbar_label=labels[
                                                axes_parameters[i]],
                                            channel=channel,
                                            parameter=axes_parameters[i],
                                            ax=axes[i], groundscatter=grndflg,
                                            cmap=cmap[axes_parameters[i]],
                                            zmin=boundary_ranges[
                                                axes_parameters[i]][0],
                                            zmax=boundary_ranges[
                                                axes_parameters[i]][1],
                                            yspacing=5,
                                            background=background,
                                            range_estimation=range_estimation,
                                            coords=coords, latlon=latlon,
                                            **kwargs)
                cbar = rt_rtn['cb']
                x = rt_rtn['data']['x']
                # Overwriting velocity ticks to get a better pleasing
                # look on the colorbar
                # Preference by Marina Schmidt
                if axes_parameters[i] == 'v':
                    locator = ticker.LinearLocator(numticks=5)
                    ticks =\
                        locator.\
                        tick_values(vmin=boundary_ranges[
                                        axes_parameters[i]][0],
                                    vmax=boundary_ranges[
                                        axes_parameters[i]][1])
                    if ticks[0] < boundary_ranges[axes_parameters[i]][0]:
                        ticks[0] = boundary_ranges[axes_parameters[i]][0]

                    if ticks[-1] > boundary_ranges[axes_parameters[i]][1]:
                        ticks[-1] = boundary_ranges[axes_parameters[i]][1]
                    cbar.set_ticks(ticks)
                if latlon == 'lat' and coords == Coords.AACGM:
                    axes[i].set_ylabel('Mag Latitude ($^\circ$)')
                elif latlon == 'lon' and coords == Coords.AACGM:
                    axes[i].set_ylabel('Mag Longitude ($^\circ$)')
                elif latlon == 'lat' and coords == Coords.GEOGRAPHIC:
                    axes[i].set_ylabel('Geo Latitude ($^\circ$)')
                elif latlon == 'lon' and coords == Coords.GEOGRAPHIC:
                    axes[i].set_ylabel('Geo Longitude ($^\circ$)')
                elif range_estimation == RangeEstimation.SLANT_RANGE:
                    axes[i].set_ylabel('Slant Range (km)')
                elif range_estimation == RangeEstimation.GSMR:
                    axes[i].set_ylabel('Ground Scatter\nMapped Range\n(km)')
                elif range_estimation == RangeEstimation.GSMR_BRISTOW:
                    axes[i].set_ylabel('Ground Scatter\nMapped Range\n(km)')
                elif range_estimation == RangeEstimation.HALF_SLANT:
                    axes[i].set_ylabel('Slant Range/2\n(km)')
                elif range_estimation == RangeEstimation.RANGE_GATE:
                    axes[i].set_ylabel('Range Gates')
                else:
                    axes[i].set_ylabel('Time of Flight\n(ms)')
            if i < num_plots-1:
                axes[i].set_xticklabels([])
            # last plot needs the label on the x-axis
            else:
                axes[i].set_xlabel('Date (UTC)')

        if title is None:
            plt.title(cls.__generate_title(x[0], x[-1], beam_num,
                                           channel), y=2.4)
        else:
            plt.title(title, y=2.4)
        plt.subplots_adjust(wspace=0, hspace=0)
        if watermark:
            fig.text(0.90, 0.99, "Not for Publication Use", fontsize=75,
                     color='gray', ha='right', va='top',
                     rotation=-38, alpha=0.3)

        return {'ax': axes,
                'ccrs': None,
                'cm': cmap,
                'cb': None,
                'fig': fig,
                'data': 'Individual range-time plots will return full data.'
                }


    @classmethod
    def __generate_title(cls, start_time: datetime, end_time: datetime,
                         beam_num: int, channel: int) -> str:
        if cls.dmap_data[0]['fitacf.revision.major'] == 5:
            version = "2.5"
        else:
            version = "{major}.{minor}"\
                    "".format(major=cls.dmap_data[0]['fitacf.revision.major'],
                              minor=cls.dmap_data[0]['fitacf.revision.minor'])
        if cls.dmap_data[0]['origin.code'] == 100:
            radar_system = " (Borealis)"
        else:
            # I would put ROS but Alaska and AGILE DARN might have their own
            # systems and not sure how to decipher between them. If something
            # changes in the file structure, then I can add it here.
            radar_system = ""
        radar_name = SuperDARNRadars.radars[RadarID(cls.dmap_data[0]['stid'])].name
        # Date time formats:
        #   %Y - year
        #   %b - month abbreviation
        #   %d - day
        #   %H - Hour
        #   %M - 2-digit minutes
        if end_time.day == start_time.day:
            end_format = "%H:%M"
        elif end_time.month == start_time.month:
            end_format = "%d %H:%M"
        elif end_time.year == start_time.year:
            end_format = "%b %d %H:%M"
        else:
            end_format = "%Y %b %d %H:%M"
        title_format =\
            "{name}{system} Fitacf {version}"\
            "{system} {start_date} - {end_date}  Beam {num}"\
            "".format(name=radar_name, version=version,
                      system=radar_system,
                      start_date=start_time.strftime("%Y %b %d %H:%M"),
                      end_date=end_time.strftime(end_format),
                      num=beam_num)
        if type(channel) is int:
            title_format += " channel {ch_num}".format(ch_num=channel)
        return title_format

    @classmethod
    # TODO: could parallelize this method
    def __filter_data_check(cls, dmap_record: dict,
                            settings: dict, j: int) -> bool:
        """
        checks for data that does not meet the criteria of the filtered
        settings

        Parameters
        ----------
        dmap_record : dict
            dictionary of the dmap record fields
        settings : dict
            dictionary of the settings list
        j : int
            index on the slist or range gate to check the array values

        Returns
        -------
        pass_flg : bool
            boolean indicating if the dmap_record passes all filter checks
        """
        pass_flg = True
        for key, value in settings['min_array_filter'].items():
            if dmap_record[key][j] < value:
                pass_flg = False
                break
        for key, value in settings['max_array_filter'].items():
            if dmap_record[key][j] > value:
                pass_flg = False
                break

        for key, value in settings['min_scalar_filter'].items():
            if dmap_record[key] < value:
                pass_flg = False
                break

        for key, value in settings['max_scalar_filter'].items():
            if dmap_record[key] > value:
                pass_flg = False
                break

        for key, value in settings['equal_scalar_filter'].items():
            if dmap_record[key] != value:
                pass_flg = False
                break

        return pass_flg



    @classmethod
    def plot_coord_time(cls, dmap_data: List[dict], parameter: str = 'v',
                        beam_num: int = 0, channel: int = 'all', ax=None,
                        background: str = 'w', background_alpha: float = 1.0,
                        groundscatter: bool = False,
                        zmin: int = None, zmax: int = None,
                        coords: object = Coords.AACGM, latlon: str = 'lat',
                        start_time: datetime = None, end_time: datetime = None,
                        colorbar: plt.colorbar = None, ymin: int = None,
                        ymax: int = None, yspacing: int = 2,
                        range_estimation: RangeEstimation =
                        RangeEstimation.SLANT_RANGE,
                        colorbar_label: str = '',
                        norm=colors.Normalize, cmap: str = None,
                        filter_settings: dict = {},
                        date_fmt: str = '%y/%m/%d\n %H:%M',
                        round_start: bool = True,
                        plot_equatorward: bool = False, **kwargs):
        """
        Plots a range-time parameter plot of the given
        field name in the dmap_data using coordinates in latitude and
        longitude for the y-axis

        Future Work
        -----------
        Support for other data input, like "time"
        dictionary key containing the datetime. However,
        further discussion is needed if this will be the keys
        name or maybe another input.

        Parameters
        -----------
        dmap_data: List[dict]
        parameter: str
            key name indicating which parameter to plot.
            Default: v (Velocity)
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The channel 0, 1, 2, 'all'
            Default : 'all'
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        groundscatter : boolean or str
            Flag to indicate if groundscatter should be plotted. If string
            groundscatter will be represented by that color else grey.
            Default : False
        background : str
            color of the background in the plot
            default: white
        background_alpha : float
            alpha (transparency) of the background in the plot
            default: 1.0 (opaque)
        zmin: int
            Minimum normalized value
            Default: minimum parameter value in the data set
        zmax: int
            Maximum normalized value
            Default: maximum parameter value in the data set
        coords: Coords
            set default coordinate system for the coordinate time plot
            default: Coords.AACGM
        latlon: str
            set which coordinate in the system you want to plot with
            default: latitude 'lat'
        ymax: int
            Sets the maximum y value
            Default: None, uses 'nrang' from data
        ymin: int
            sets the minimum y value
            Default: None, uses 0 range gate or minimum slant-range value
        yspacing: int
            sets the spacing between ticks
            Default: 2
        range_estimation: RangeEstimation
            set the y-axis to a desired range estimation calculation
            Default: RangeEstimation.SLANT_RANGE
        norm: matplotlib.colors.Normalization object
            This object use dependency injection to use any normalization
            method with the zmin and zmax.
            Default: colors.Normalization()
        start_time: datetime
            Start time of the plot x-axis as a datetime object
            Default: rounded to nearest hour-30 minutes from the first
                     record containing the chosen parameters data
        end_time: datetime
            End time of the plot x-axis as a datetime object
            Default: last record of the chosen parameters data
        date_fmt : str
            format of x-axis date ticks, follow datetime format
            Default: '%y/%m/%d\n %H:%M' (Year/Month/Day Hour:Minute)
        colorbar: matplotlib.pyplot.colorbar
            Setting a predefined colorbar for the range-time plot
            If None, then a colorbar will be created for the plot
            Default: None
        colorbar_label: str
            the label that appears next to the color bar
            Default: ''
        cmap: str or matplotlib.cm
            matplotlib colour map
            https://matplotlib.org/tutorials/colors/colormaps.html
            Default: PyDARNColormaps.PYDARN_VELOCITY
            note: to reverse the color just add _r to the string name
        filter_settings: dict
            dictionary of the following keys for filtering data out:
            max_array_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out any data points
                that is above this value.
            min_array_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out any data points that is
                below this value.
            max_scalar_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out data sections that is
                above this value.
            min_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is below this value.
            equal_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is does not equal the value.
        round_start: bool=True
                option to round the start time to give tick at start of xaxis
                Set True to round, set False to plot from start of data.
                Default: True
        plot_equatorward: bool = False
                option if the radar plots data in latitude that changes
                direction to plot the equator-ward or pole-ward data
                No option to overplot.
                Default: False (plot poleward data only)
        kwargs:
            used for other methods in pyDARN
                - reflection_height
        Raises
        ------
        UnknownParameterError
        IncorrectPlotMethodError
        NoDataFoundError
        IndexError

        Returns
        -------
        im: matplotlib.pyplot.pcolormesh
            matplotlib object from pcolormesh
        cb: matplotlib.colorbar
            matplotlib color bar
        cmap: matplotlib.cm
            matplotlib color map object
        time_axis: list
            list representing the x-axis datetime objects
        y_axis: list
            list representing the y-axis range gates
        z_data: 2D numpy array
            2D array of the parameters values at the given time and range gate

        See Also
        ---------
        colors: https://matplotlib.org/2.0.2/api/colors_api.html
        color maps: PyDARNColormaps or
                    https://matplotlib.org/tutorials/colors/colormaps.html
        normalize:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.colors.Normalize.html
        colorbar:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.colorbar.html
        pcolormesh:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.pcolormesh.html

        """
        # Settings
        plot_filter = {'min_array_filter': dict(),
                       'max_array_filter': dict(),
                       'min_scalar_filter': dict(),
                       'max_scalar_filter': dict(),
                       'equal_scalar_filter': dict()}

        plot_filter.update(filter_settings)

        # If an axes object is not passed in then store
        # the equivalent object in matplotlib. This allows
        # for variant matplotlib plotting styles.
        if not ax:
            ax = plt.gca()

        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            index_first_match = next(i for i, d in enumerate(dmap_data)
                                     if parameter in d)
        except StopIteration:
            raise plot_exceptions.UnknownParameterError(parameter)
        cls.dmap_data = dmap_data
        check_data_type(cls.dmap_data, parameter, 'array', index_first_match)
        start_time, end_time = cls.__determine_start_end_time(start_time,
                                                              end_time)

        # y-axis coordinates, i.e., range gates,
        # TODO: implement variant other coordinate systems for the y-axis
        # y shape needs to be +1 longer as requirement of how pcolormesh
        # draws the pixels on the grid

        # because nrang can change based on mode we need to look
        # for the largest value
        y_max = max(record['nrang'] for record in cls.dmap_data)
        y = np.arange(0, y_max+1, 1)

        # z: parameter data mapped into the color mesh
        z = np.zeros((1, y_max)) * np.nan

        # x: time date data
        x = []

        # We cannot simply use numpy's built in min and max function
        # because of the groundscatter value :(

        # These flags indicate if zmin and zmax should change
        set_zmin = True
        set_zmax = True
        if zmin is None:
            zmin = cls.dmap_data[index_first_match][parameter][0]
            set_zmin = False
        if zmax is None:
            zmax = cls.dmap_data[index_first_match][parameter][0]
            set_zmax = False

        for dmap_record in cls.dmap_data:
            # get time difference to test if there is some gap data
            rec_time = time2datetime(dmap_record)
            diff_time = 0.0
            if rec_time > end_time:
                break
            if x != []:
                # 60.0 seconds in a minute
                delta_diff_time = abs(rec_time - x[-1])
                diff_time = delta_diff_time.seconds/60.0
                # Abs added above as some files have data out of order
                # abs stops the code from hanging and plotting over a day
                # of white space, but user needs to be warned that the
                # output may be incorrect
                if (rec_time - x[-1]) < timedelta(0):
                    # warnings are turned off for summary plots so
                    # for now print to console
                    # May be repeated, but will show what records are out
                    # of time order by doing so to help user
                    print("Please be aware that the data for timestamp {}"
                          " contains a record that is not"
                          " in time order. As such the plot of the"
                          " data may not be correct, you can solve"
                          " this by sorting the data stream by date"
                          " before plotting.".format(rec_time))


            # separation roughly 2 minutes
            if diff_time > 2.0:
                # if there is gap data (no data recorded past 2 minutes)
                # then fill it in with white space
                for _ in range(0, int(np.floor(diff_time/2.0))):
                    x.append(x[-1] + timedelta(0, 120))
                    i = len(x) - 1  # offset since we start at 0 not 1
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros([1, y_max]) * np.nan,
                                      axis=0)
            # Get data for the provided beam number
            if (beam_num == 'all' or dmap_record['bmnum'] == beam_num) and\
               (channel == 'all' or
                    dmap_record['channel'] == channel):
                if start_time <= rec_time:
                    # construct the x-axis array
                    # Numpy datetime is used because it properly formats on the
                    # x-axis
                    x.append(rec_time)
                    # I do this to avoid having an extra loop to just count how
                    # many records contain the beam number
                    i = len(x) - 1  # offset since we start at 0 not 1

                    # insert a new column into the z_data
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros([1, y_max]) * np.nan,
                                      axis=0)
                    try:
                        if len(dmap_record[parameter]) == dmap_record['nrang']:
                            good_gates = range(len(dmap_record[parameter]))
                        else:
                            good_gates = dmap_record['slist']

                        # get the range gates that have "good" data in it
                        for j in range(len(good_gates)):
                            # if it is groundscatter store a very
                            # low number in that cell
                            if groundscatter and\
                               dmap_record['gflg'][j] == 1:
                                # chosen value from davitpy to make the
                                # groundscatter a different color
                                # from the color map
                                z[i][good_gates[j]] = -1000000
                            # otherwise store parameter value
                            # TODO: refactor and clean up this code
                            elif cls.__filter_data_check(dmap_record,
                                                         plot_filter, j):
                                z[i][good_gates[j]] = \
                                        dmap_record[parameter][j]
                                # calculate min and max value
                                if not set_zmin and\
                                   z[i][good_gates[j]] < zmin:
                                    zmin = z[i][good_gates[j]]
                                if not set_zmax and \
                                   z[i][good_gates[j]] > zmax:
                                    zmax = z[i][good_gates[j]]
                    # a KeyError may be thrown because slist is not created
                    # due to bad quality data.
                    except KeyError:
                        continue
        x.append(end_time)
        # Check if there is any data to plot
        if np.all(np.isnan(z)):
            raise plot_exceptions.\
                    NoDataFoundError(parameter, beam_num,
                                     start_time=start_time,
                                     end_time=end_time,
                                     opt_beam_num=cls.dmap_data[0]['bmnum'])

        frang = int(cls.dmap_data[0]['frang'])
        rsep = int(cls.dmap_data[0]['rsep'])

        # MLT is not applicable for plotting
        if coords == Coords.AACGM_MLT:
            raise Exception("Error: MLT cannot be used in coord-time plots. "
                            "Please choose Coords from AACGM or GEOGRAPHIC.")
        # Get position of the range gates in lat lon
        lats,lons = coords(stid=RadarID(dmap_data[0]['stid']), rsep=rsep, frang=frang,
                           date=start_time, center=False,
                           gates=[0, dmap_data[0]['nrang']],
                           range_estimation=range_estimation, **kwargs)

        if latlon == 'lat':
            y = lats[:, beam_num]
            # If the FOV is over the pole, only plot up to the pole to avoid
            # overplotting data unless the user specifies downward
            if y[0] > 0:
                diff_y = [j-i for i, j in zip(y[:-1], y[1:])]
            else:
                # If southern hemisphere, just switch diff_y values to get
                # correct position
                diff_y = [-(j-i) for i, j in zip(y[:-1], y[1:])]

            if any(x < 0 for x in diff_y) and not plot_equatorward:
                yind = np.min([i for i in range(len(diff_y)) if diff_y[i] < 0]) - 1
                warnings.warn('Warning: This radar has a field of view that over'
                            'plots data in latitude. Only the partial '
                            'field of view to range gate '
                            '{} is plotted. To see the other section of data '
                            'use the keyword plot_equatorward=True.'
                            .format(yind,))
                y = y[:yind]
                y = np.append(y, y[-1])
                z = z[:, :yind]
            if any(x < 0 for x in diff_y) and plot_equatorward:
                yind = np.min([i for i in range(len(diff_y)) if diff_y[i] < 0]) -1
                warnings.warn('Warning: This radar has a field of view that over'
                            'plots data in latitude. Only the partial '
                            'field of view from range gate '
                            '{} is plotted. To see the other section of data '
                            'use the keyword plot_equatorward=False.'
                            .format(yind))
                y = y[yind-1:]
                z = z[:, yind-1:]
        elif latlon == 'lon':
            y = lons[:, beam_num]
        else:
            raise Exception('Error: latlon values can be "lat" or "lon" only.')

        time_axis, y_axis = np.meshgrid(x, y)
        z_data = np.ma.masked_where(np.isnan(z.T), z.T)
        Default = {'noise.sky': (1e0, 1e5),
                   'tfreq': (8, 22),
                   'nave': (0, 60),
                   'p_l': (0, 45),
                   'v': (-200, 200),
                   'w_l': (0, 250),
                   'elv': (0, 45)}

        if np.isinf(zmin) and zmin < 0:
            zmin = Default[parameter][0]
            warnings.warn("Warning: zmin is -inf, set zmin to {}. You can"
                          "set zmin and zmax in the function"
                          " options".format(zmin))
        if np.isinf(zmax) and zmax > 0:
            zmax = Default[parameter][1]
            warnings.warn("Warning: zmax is inf, set zmax to {}. You can"
                          "set zmin and zmax in the functions"
                          " options".format(zmax))
        norm = norm(zmin, zmax)
        if isinstance(cmap, str):
            cmap = colormaps.get_cmap(cmap)
        else:
            # need to do this as matplotlib 3.5 will
            # not all direct mutations of the object
            cmaps = {'p_l': copy.copy(colormaps.get_cmap('plasma')),
                     'v': PyDARNColormaps.PYDARN_VELOCITY,
                     'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                     'elv': PyDARNColormaps.PYDARN}
            cmap = cmaps[parameter]

        # set the background color, this needs to happen to avoid
        # the overlapping problem that occurs
        cmap.set_bad(color=background, alpha=background_alpha)
        # plot!
        im = ax.pcolormesh(time_axis, y_axis, z_data, lw=0.01,
                           cmap=cmap, norm=norm, **kwargs)

        if isinstance(groundscatter, str):
            ground_scatter = np.ma.masked_where(z_data != -1000000, z_data)
            gs_color = colors.ListedColormap([groundscatter])
            ax.pcolormesh(time_axis, y_axis, ground_scatter, lw=0.01,
                          cmap=gs_color, norm=norm, **kwargs)

        elif groundscatter:
            ground_scatter = np.ma.masked_where(z_data != -1000000, z_data)
            gs_color = colors.ListedColormap(['grey'])
            ax.pcolormesh(time_axis, y_axis, ground_scatter, lw=0.01,
                          cmap=gs_color, norm=norm, **kwargs)

        # setup some standard axis information
        if ymax is None:
            ymax = np.max(y)

        if ymin is None:
            ymin = np.min(y)

        ax.set_ylim(ymin, ymax)
        if yspacing != 2:
            ax.yaxis.set_ticks(np.arange(np.floor(ymin), np.ceil(ymax),
                                         yspacing))
        elif ymax-ymin < 100:
            ax.yaxis.set_ticks(np.arange(np.floor(ymin), np.ceil(ymax),
                                         yspacing))
        else:
            ax.yaxis.set_ticks(np.arange(np.floor(ymin/10)*10,
                                         np.ceil(ymax/10)*10,
                                         yspacing*5))

        # SuperDARN file typically are in 2hr or 24 hr files
        # to make the minute ticks sensible, the time length is detected
        # then a interval is picked. 30 minute ticks for 24 hr plots
        # and 5 minute ticks for 2 hour plots.
        data_time_length = end_time - start_time
        # 3 hours * 60 minutes * 60 seconds
        if data_time_length.total_seconds() > 3*60*60:
            tick_interval = 30
        else:
            tick_interval = 1
        # byminute keyword makes sure that the ticks are situated at
        # the minute or half hour marks, rather than at a set interval
        ax.xaxis.set_minor_locator(
            dates.MinuteLocator(byminute=range(0, 60, tick_interval)))

        # Upon request of Daniel Billet and others, I am rounding
        # the time down so the plotting x-axis will show the origin
        # time label
        # Updated to give option to round down and make sure
        # rounding to same frequency as plot axis ticks if less than 1 hour
        if round_start:
            major_locator, _ = plt.xticks()
            dt = dates.num2date(major_locator[1]) -\
                dates.num2date(major_locator[0])
            tick_sep = dt.seconds//60
            if tick_sep > 0:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % tick_sep,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
            else:
                rounded_down_start_time = x[0] -\
                    timedelta(minutes=x[0].minute % 15,
                              seconds=x[0].second,
                              microseconds=x[0].microsecond)
        else:
            rounded_down_start_time = x[0]

        ax.set_xlim([rounded_down_start_time, x[-1]])
        ax.xaxis.set_major_formatter(dates.DateFormatter(date_fmt))
        # Make sure the axis flips with pole at the top for southern data
        if y[0] < 0 and latlon == 'lat':
            ax.invert_yaxis()

        if range_estimation != RangeEstimation.RANGE_GATE:
            ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(2))
        else:
            ax.yaxis.set_minor_locator(ticker.MultipleLocator(5))
        # so the plots gets to the ends
        ax.margins(0)

        # Create color bar if None supplied
        if not colorbar:
            with warnings.catch_warnings():
                warnings.filterwarnings('error')
                try:
                    if isinstance(norm, colors.LogNorm):
                        if zmin == 0:
                            cb = ax.figure.colorbar(im, ax=ax, extend='max')
                        else:
                            cb = ax.figure.colorbar(im, ax=ax, extend='both')
                    else:
                        locator = ticker.MaxNLocator(symmetric=True,
                                                     min_n_ticks=3,
                                                     integer=True,
                                                     nbins='auto')
                        ticks = locator.tick_values(vmin=zmin, vmax=zmax)
                        if zmin == 0:
                            cb = ax.figure.colorbar(im, ax=ax, extend='max',
                                                    ticks=ticks)
                        else:
                            cb = ax.figure.colorbar(im, ax=ax, extend='both',
                                                    ticks=ticks)

                except (ZeroDivisionError, Warning):
                    raise rtp_exceptions.RTPZeroError(parameter, beam_num,
                                                      zmin, zmax,
                                                      norm) from None
        else:
            cb = colorbar
        if colorbar_label != '':
            cb.set_label(colorbar_label)
        return {'ax': ax,
                'ccrs': None,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'plot_data': im,
                         'x': x,
                         'y': y,
                         'z': z_data}
                }


    # TODO: if used in other plotting methods then this should moved to
    #       utils
    @classmethod
    def __determine_start_end_time(cls, start_time: datetime,
                                   end_time: datetime) -> tuple:
        """
        Sets the start and end time based on import of dmap_data

        Parameter
        ---------
        start_time: datetime
            Start time is used to check if it was set or not
        end_time: datetime
            End time is used to check if it was set or not

        Returns
        -------
        start_time: datetime
        end_time: datetime
        """
        if not start_time:
            start_time = time2datetime(cls.dmap_data[0])
        if not end_time:
            end_time = time2datetime(cls.dmap_data[-1])
        return start_time, end_time



# (C) Copyright 2021 SuperDARN Canada, University of Saskatachewan
# Author(s): Marina Schmidt
# (C) Copyright 2021 University of Scranton
# Author(s): Francis Tholley
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
#  2021-09-15 Francis Tholley moved the chisham and standard virtual
#  height models to separate file for better encapsulation/modularity
#  2022-03-04 Marina Schmidt add the VH_Types class to the bottom
""" virtual_heights.py comprises of different of virtual height models"""
import enum

def chisham(target_range: float, **kwargs):
    """
    Mapping ionospheric backscatter measured by the SuperDARN HF
    radars – Part 1: A new empirical virtual height model by
    G. Chisham 2008 (https://doi.org/10.5194/angeo-26-823-2008)

    Parameters
    ----------
    target_range: float
        is the range from radar to the target (echos)
        sometimes known as slant range [km]
    kwargs: is only needed to avoid key item errors

    Returns
    -------
    altered target_range (slant range) [km]
    """
    # Model constants
    A_const = (108.974, 384.416, 1098.28)
    B_const = (0.0191271, -0.178640, -0.354557)
    C_const = (6.68283e-5, 1.81405e-4, 9.39961e-5)

    # determine which region of ionosphere the gate
    if target_range < 115:
        return (target_range / 115.0) * 112.0
    elif target_range < 787.5:
        return A_const[0] + B_const[0] * target_range + C_const[0] *\
                 target_range**2
    elif target_range <= 2137.5:
        return A_const[1] + B_const[1] * target_range + C_const[1] *\
                 target_range**2
    else:
        return A_const[2] + B_const[2] * target_range + C_const[2] *\
                 target_range**2


def standard_virtual_height(target_range: float, cell_height: int = 300,
                            **kwargs):
    """
    cell_height, target_range and x_height are in km
    Default values set in virtual height model described
    Mapping ionospheric backscatter measured by the SuperDARN HF
    radars – Part 1: A new empirical virtual height model by
    G. Chisham 2008
    Equation (1) in the paper
    < 150 km climbing into the E region
    150 - 600 km E region scatter
    (Note in the paper 400 km is the edge of the E region)
    600 - 800 km is F region

    Parameters
    ----------
    target_range: float
        is the range from radar to the target (echos)
        sometimes known as slant range [km]
    cell_height: int
        the default height of the echo if the target_range
        is within a certain range
    kwargs: is only needed to avoid key item errors

    Returns
    -------
    altered target_range (slant range) [km]
    """
    # TODO: why 115?
    # map everything into the E region
    if cell_height <= 150 and target_range > 150:
        return cell_height
    # virtual height equation (1) from the above paper
    elif target_range < 150:
        return (target_range / 150.0) * 115
    elif target_range >= 150 and target_range <= 600:
        return 115
    elif target_range > 600 and target_range < 800:
        return (target_range - 600) / 200 * (cell_height - 115) + 115
    # higher than 800 km
    else:
        return cell_height


class VHModels(enum.Enum):
    """
    This virtual height models class is to list the current
    virtual height model user can pick from

    enumerators:
        STANDARD: Standard_Virtual_height (km)
        CHISHAM: chisham (km)
    """

    STANDARD = (standard_virtual_height, )
    CHISHAM = (chisham, )

    # Need this to make the functions callable
    def __call__(self, *args, **kwargs):
        return self.value[0](*args, **kwargs)



# copyright (C) SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
#
# Modifications:
#
# Disclaimer: pyDARN is licensed under the GPL v 3.0 found in LICENSE
#
import warnings

def nightshade_warning():
    """
    This warning prints when a user tries to use something that is not
    implemented yet, however the plot will still plot without the
    feature
    """
    warnings.warn("Nightshade is not implemented for plots in AACGMv2 "
                  "coordinates. You can plot nightshade in geographic "
                  "coordinates, or you can call the terminator function "
                  "to return the antisolarposition and radius to "
                  "decide how you would like to plot."
                  "https://pydarn.readthedocs.io/en/main/user/coordinates/")

def partial_record_warning():
    """
    prints a warning that the data chosen to be plotted is missing some
    keys that may produce a blank/half a plot
    """
    warnings.warn("Please be aware that the data chosen to be plotted"
                  " contains a partial record. As such the plot may"
                  " be empty or only partially filled.")

def standard_warning_format(message: str, category: str, filename: str,
                            lineno: int, file: str = None,
                            line: int = None) -> str:
    """
    Sets the standard warning message format to be:
        filename: lineno: category: message

    Parameters
    ----------
        message: str
            warning message to be printed to a user
        category: str
            type of the warning message
        filename: str
            name of the associated to the warning message
        lineno : int
            line number where the warning message was raised
        file : str
        line : str

    Returns
    -------
        formatted warning message to be printed to the console
    """
    return "{filename}: {linenum}: {category}:"\
           " {message}\n".format(filename=filename,
                                 linenum=lineno,
                                 category=category.__name__,
                                 message=message)


def only_message_warning_format(message: str, category: str, filename: str,
                                lineno: int, file: str = None,
                                line: int = None) -> str:
    """
    Sets the standard warning message format to be:
        filename: lineno: category: message

    Parameters
    ----------
        message: str
            warning message to be printed to a user
        category: str
            type of the warning message
        filename: str
            name of the associated to the warning message
        lineno : int
            line number where the warning message was raised
        file : str
        line : str

    Returns
    -------
        formatted warning message to be printed to the console
    """
    return "{filename} {linenum} {category}: {message}\n"\
        "".format(filename=filename, linenum=lineno,
                  category=category.__name__, message=message)



#Copyright 2018 SuperDARN Canada, Univeristy Saskatchewan
#Author(s): Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
# 2022-03-10 MTS - removed radar_fov from the __init__ file
# 2023-06-20 PXP - added TimeSeriesParams to the __init__ file
"""
Init file to setup the logging configuration and linking pyDARN's
module, classes, and functions.
"""
# KEEP THIS FILE AS MINIMAL AS POSSIBLE!
# ruff: noqa: F401
# version file
from .version import __version__

# Import io for pyDARN
from .io.superdarn_io import SuperDARNRead

# Importing pydarn exception classes
from .exceptions import rtp_exceptions, plot_exceptions, radar_exceptions
from .exceptions.warning_formatting import (standard_warning_format,
    only_message_warning_format, partial_record_warning, nightshade_warning)

# importing utils
from .utils.constants import Re, EARTH_EQUATORIAL_RADIUS, C
from .utils.citations import Citations
from .utils.range_estimations import RangeEstimation
from .utils.virtual_heights import VHModels
from .utils.conversions import dmap2dict
from .utils.plotting import (MapParams, TimeSeriesParams, check_data_type,
    time2datetime, find_record, determine_embargo, add_embargo)
from .utils.general_utils import GeneralUtils
from .utils.superdarn_radars import RadarID, SuperDARNRadars
from .utils.superdarn_cpid import SuperDARNCpids
from .utils.superdarn_radars import Hemisphere, read_hdw_file, get_hdw_files
from .utils.scan import find_records_by_datetime, find_records_by_scan
from .utils.geo import geocentric_coordinates, calculate_azimuth
from .utils.coordinates import Coords
from .utils.terminator import terminator
from .utils.recalculate_elevation import recalculate_elevation
from .utils.filters import Boxcar

# import plotting
from .plotting.color_maps import PyDARNColormaps
from .plotting.projections import Projs
from .plotting.rtp import RTP
from .plotting.fan import Fan
from .plotting.grid import Grid
from .plotting.acf import ACF
from .plotting.power import Power
from .plotting.maps import Maps
from .plotting.iq import IQ



# (C) Copyright SuperDARN Canada, University of Saskatchewan 2022
# Author: Carley Martin 20221101
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# About:
# Modified from the RST elevation_v2.c code (Simon Shepherd's code).
# Method: https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2017RS006348
# THIS CODE IS DESIGNED TO AMEND ELEVATION
# AFTER FITACF FITTING (AT THE VISUALIZATION STEP TO QUICKLY CHECK THE EFFECTS
# OF DIFFERENT TDIFF VALUES) NOT TO REPLACE THE FITTING ITSELF
#
# Modifications:
# 20221221 - Bharat Kunduri: Updated to RST elevation code

import numpy as np

from typing import List
from copy import deepcopy

from pydarn import (SuperDARNRadars, C, RadarID)


def recalculate_elevation(dmap_data: List[dict], tdiff: float,
                          overwrite: bool = False,
                          interferometer_offset: list = None):
    """
    Recalculates elevation values for a given tdiff Value

    Parameters
    -----------
    dmap_data: list of dictionaries
        fitacf data
    tdiff: float
        propagation time from interferometer array to phasing matrix
        input minus propagation time from main array antenna, microseconds
    overwrite: bool
        If true then return a new dmap_data with new elevation to plot with
        if false then return dictionary of new elevations for further use
    interferometer_offset: list
        select position of interferometer array wrt the main array
        needs to be list of [X, Y, Z] e.g. [0.0, 100.0, 1.0]

    Raises
    ------

    Returns
    -------
    elv_amended: dictionary of lists
        amended elevation values for each record given
    """
    if not overwrite:
        # Make phi1 output into dictionary
        elv_amended = {}
    else:
        dmap_amended = deepcopy(dmap_data)

    # Hardware config for radar
    # Doesn't have to be in the loop, since we're accessing only the first rec
    radar_hdw = SuperDARNRadars.radars[RadarID(dmap_data[0]['stid'])].hardware_info
    if interferometer_offset is not None:
        int_pos = interferometer_offset
    else:
        int_pos = radar_hdw.interferometer_offset

    # If inteferometer is infront (+1) or behind (-1) main array
    if int_pos[1] < 0:
        sgn = -1.0
    else:
        sgn = 1.0

    boff = (radar_hdw.beams / 2.0) - 0.5

    # For each record in data, recalculate the elevation
    for ind in range(0, len(dmap_data)):
        # If there is no elevation data, append empty list and skip
        if 'phi0' not in dmap_data[ind]:
            print("No elevation data. 'phi0' parameter missing"
                  " from the record")
            if not overwrite:
                elv_amended[ind] = []
            continue

        # Beam direction off boresight in RADIANS
        phi0 = np.radians(radar_hdw.beam_separation
                          * (dmap_data[ind]['bmnum'] - boff))
        # Cos and Sin of phi in shape of phi0
        cp0 = np.cos(phi0)
        sp0 = np.sin(phi0)

        # Phase delay [radians] due to electrical path difference.
        psi_ele = (-2.0 * np.pi * (dmap_data[ind]['tfreq'])
                   * 1000.0 * tdiff * 1.0e-6)
        # Elevation angle (a0) where psi (phase difference) is maximum
        a0 = np.arcsin(sgn * int_pos[2] * cp0
                       / np.sqrt(int_pos[1]**2 + int_pos[2]**2))
        if a0 < 0:
            a0 = 0
        ca0 = np.cos(a0)
        sa0 = np.sin(a0)

        # maximum phase = psi_ele + psi_geo(a0)
        psi_max = psi_ele + 2.0 * np.pi * (dmap_data[ind]['tfreq']) *\
            (1e3 / C) * (int_pos[0] * sp0 + int_pos[1]
                         * np.sqrt(ca0*ca0 - sp0*sp0) + int_pos[2] * sa0)

        # compute the number of 2pi factors necessary to map to correct region
        dpsi = (psi_max - dmap_data[ind]['phi0'])
        if int_pos[1] > 0:
            n2pi = np.floor(dpsi / (2.0 * np.pi))
        else:
            n2pi = np.ceil(dpsi / (2.0 * np.pi))
        d2pi = n2pi * 2.0 * np.pi
        # map observed phase to correct extended phase
        psi_obs = dmap_data[ind]['phi0'] + d2pi
        # solve for the elevation angle
        E = (psi_obs / (2.0*np.pi*dmap_data[ind]['tfreq']*1.0e3)
             + tdiff*1e-6) * C - int_pos[0] * sp0

        alpha = np.arcsin((E*int_pos[2]
                           + np.sqrt(E*E * int_pos[2]**2
                           - (int_pos[1]**2 + int_pos[2]**2)
                           * (E*E - int_pos[1]*int_pos[1]*cp0*cp0)))
                          / (int_pos[1]*int_pos[1] + int_pos[2]*int_pos[2]))

        # Convert theta back to degrees
        if overwrite:
            dmap_amended[ind].update({'elv': np.degrees(alpha)})
        else:
            elv_amended[ind] = np.degrees(alpha)

    if overwrite:
        return dmap_amended
    else:
        return elv_amended



# Copyright (C) 2023 SuperDARN Canada, University of Saskatchewan
# Author: Shibaji Chakraborty
#
# Modifications:
# 20230202 - CJM: Integrate code into pyDARN
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.


"""filters.py: Module is dedicated to filters used on FITACF SuperDARN data."""

import copy
import datetime as dt
import numpy as np
import warnings

from pydarn import standard_warning_format

warnings.formatwarning = standard_warning_format

class Gate(object):
    """Class object to hold each range cell value

    Methods
    -------
    NA
    """
    def __init__(self, bm, i, params=["v", "w_l", "gflg", "p_l", "v_e", "elv"],
                 gflg_type=-1):
        """
        Initialize the parameters which will be stored in Gate
        Parameters
        ----------
        bm : beam object
            A collection of data from the filtering process
        i : int
            Index to store in collection
        params :
            Parameters to store in collection
        Returns
        -------
        NA
        """
        for p in params:
            if len(getattr(bm, p)) > i:
                setattr(self, p, getattr(bm, p)[i])
            else:
                setattr(self, p, np.nan)
        if gflg_type >= 0 and len(getattr(bm, "gsflg")[gflg_type]) > 0:
            setattr(self, "gflg", getattr(bm, "gsflg")[gflg_type][i])
        return


class Beam(object):
    """Class to hold one radar beam

    Methods
    -------
    set
    copy
    """
    def __init__(self):
        """ Initialize the instance """
        return

    def set(self, time, d,
            s_params=["bmnum", "noise.sky", "tfreq", "scan", "nrang"],
            v_params=["v", "w_l", "gflg", "p_l", "slist", "v_e", "elv"], k=None):
        """
        Parameters
        ----------
        time : datetime object
            Time of data in beam
        d : Dict
            Containing data for other parameters in beam
        s_params : list
            Other scalar params
        v_params : list
            Other vector params
        """
        for p in s_params:
            if p in d.keys():
                if p == "scan" and d[p] != 0:
                    setattr(self, p, 1)
                else:
                    setattr(self, p, d[p]) if k is None else setattr(self, p,
                                                                     d[p][k])
            else:
                setattr(self, p, None)
        for p in v_params:
            if p in d.keys():
                setattr(self, p, d[p])
            else:
                setattr(self, p, [])
        self.time = time
        return

    def copy(self, bm):
        """Copy all parameters"""
        for p in bm.__dict__.keys():
            setattr(self, p, getattr(bm, p))
        return


class Scan(object):
    """Class to hold one radar scans

    Methods
    -------
    update_time
    """
    def __init__(self):
        """ Initialize the instance """
        self.beams = []
        return

    def update_time(self):
        """
        Update stime and etime of the scan.
        """
        if len(self.beams) > 0:
            self.stime = min([b.time for b in self.beams])
            self.etime = max([b.time for b in self.beams])
        self.scan_time = 60 * np.rint((self.etime - self.stime)
                                      .total_seconds() / 60)
        return


class FetchData(object):
    """Class to fetch data from fitacf files

    Methods
    -------
    parse_data
    """
    def __init__(self, beam_sounds):
        """
        Initialize the variables in a fitacf file
        """
        self.beam_sounds = beam_sounds
        self.s_params = [
            "bmnum",
            "noise.sky",
            "tfreq",
            "scan",
            "nrang",
            "intt.sc",
            "intt.us",
            "mppul",
            "rsep",
            "cp",
            "frang",
            "smsep",
            "lagfr",
            "channel",
            "mplgs",
            "nave",
            "noise.search",
            "mplgexs",
            "xcf",
            "noise.mean",
            "ifmode",
            "bmazm",
            "rxrise",
            "mpinc",
        ]
        self.v_params = ["v", "w_l", "gflg", "p_l", "slist", "v_e", "elv"]
        self.scans, self.beams = [], []
        return

    def parse_data(self, by="scan"):
        """
        Parse data by data type

        Parameters
        ----------
        by: str
            "scan" - no other options for now

        Returns
        -------
        NA
        """
        for d in self.beam_sounds:
            time = dt.datetime(
                d["time.yr"],
                d["time.mo"],
                d["time.dy"],
                d["time.hr"],
                d["time.mt"],
                d["time.sc"],
                d["time.us"]
            )
            bm = Beam()
            bm.set(time, d, self.s_params, self.v_params)
            self.beams.append(bm)
        if by == "scan":
            sc = Scan()
            sc.beams.append(self.beams[0])
            for _ix, d in enumerate(self.beams[1:]):
                if d.scan == 1:
                    sc.update_time()
                    self.scans.append(sc)
                    del sc
                    sc = Scan()
                    sc.beams.append(d)
                else:
                    sc.beams.append(d)
            self.scans.append(sc)
        return


class Boxcar(object):
    """Class to filter data - Boxcar median filter.

    Methods
    -------
    format_data_for_pydarn
    run_filter
    __discard_repeating_beams__
    __do_filter__
    """
    def __init__(self, thresh=0.7, w=None, gflg_type=-1):
        """
        Initialize variables

        Parameters
        ----------
        thresh: float
            Threshold of the weight matrix
        w: list(list)
            Weight matrix
        pbnd: float
            Lower and upper bounds of IS / GS probability
        pth: float
            Probability of the threshold

        Returns
        -------
        NA
        """
        self.thresh = thresh
        if w is None:
            w = np.array(
                [
                    [[1, 2, 1], [2, 3, 2], [1, 2, 1]],
                    [[2, 3, 2], [3, 5, 3], [2, 3, 2]],
                    [[1, 2, 1], [2, 3, 2], [1, 2, 1]],
                ]
            )
        self.w = w
        self.gflg_type = gflg_type
        return

    def format_data_for_pydarn(self, original_data):
        """
        Place filtered data into original data structure for pyDARN
        only changing the data fields that were filtered

        Parameters
        ----------
        original_data: List[Dict]
            List of SuperDARN fitacf data

        Returns
        -------
        NA
        """
        # Deep copy original fitacf data
        self.copied_data = copy.deepcopy(original_data)
        # For each record in the fitacf data, find matching time in
        # filtered data, replace with the new filtered data
        for r, record in enumerate(original_data):
            # Track to see if match is found or not
            match_found = False
            record_time = dt.datetime(record["time.yr"], record["time.mo"],
                                      record["time.dy"], record["time.hr"],
                                      record["time.mt"], record["time.sc"],
                                      record["time.us"])
            for frec in self.filtered_data["beam_sounds"]:
                if record_time == frec['time']:
                    match_found = True
                    if not bool(frec['slist']):
                        # If new data is empty remove it from dictionary
                        self.copied_data[r].pop('slist', None)
                        self.copied_data[r].pop('v', None)
                        self.copied_data[r].pop('w_l', None)
                        self.copied_data[r].pop('p_l', None)
                        self.copied_data[r].pop('elv', None)
                        self.copied_data[r].pop('gflg', None)
                    else:
                        # Replace the data with new filtered data if there is
                        # new data to replace it
                        self.copied_data[r]['slist'] = np.array(frec['slist'])
                        self.copied_data[r]['v'] = np.array(frec['v'])
                        self.copied_data[r]['w_l'] = np.array(frec['w_l'])
                        self.copied_data[r]['p_l'] = np.array(frec['p_l'])
                        self.copied_data[r]['elv'] = np.array(frec['elv'])
                        self.copied_data[r]['gflg'] = np.array(frec['gflg'])
            # If no match is found for the record, then
            # empty the fields, new data needs to be empty
            if not match_found:
                self.copied_data[r].pop('slist', None)
                self.copied_data[r].pop('v', None)
                self.copied_data[r].pop('w_l', None)
                self.copied_data[r].pop('elv', None)
                self.copied_data[r].pop('gflg', None)
        return

    def run_filter(self, beam_sounds, cpus=1):
        """
        Set data and convert to scan objects

        Parameters
        ----------
        beam_sounds: List[Dict]
            List of SuperDARN fitacf data
        cpus: int
            Number of cpus available/rdesired
            to run in parallel (UNUSED)

        Returns
        -------
        copied_data: List[Dict]
            List of dictionaries that contain
            the new filtered data
        """
        warnings.warn('The boxcar filter may take 5-10 minutes to filter a ' +
                      'two hour FITACF file. For more information on this ' +
                      'filter see the documentation at ' +
                      'https://pydarn.readthedocs.io/en/main/') 
        warnings.warn('The boxcar filter may not be applicable to all data, '+
                      'for example, the boxcar filter should not be applied '+
                      'to twofsound data.')
        fd = FetchData(beam_sounds)
        fd.parse_data()
        self.scan_stacks = [fd.scans[i - 1: i + 2]
                            for i in range(1, len(fd.scans) - 1)]
        self.filtered_data = {"scans": [], "beams": [], "beam_sounds": []}
        if cpus > 1:
            # TODO: Parallel processing options
            # ray.init(num_cpus=cpus)
            # futures = [
            #    self.bx.doFilter.remote(scan_stack, comb=comb,
            #                            gflg_type=gflg_type)
            #    for scan_stack in scan_stacks
            # ]
            # self.filtered_data["scans"] = ray.get(futures)
            pass
        else:
            from collections import OrderedDict
            scans = [
                self.__do_filter__(scan_stack)
                for scan_stack in self.scan_stacks
            ]
            beams = []
            for s in scans:
                beams.extend(s.beams)
            self.filtered_data["scans"] = scans
            self.filtered_data["beams"] = beams
            self.filtered_data["beam_sounds"] = [
                OrderedDict([(k, getattr(b, k)) for k in b.__dict__.keys()])
                for b in beams
            ]
            # Format the data for pyDARN plotting and return the new
            # filtered version of the fitacf data
            self.format_data_for_pydarn(beam_sounds)
        return self.copied_data

    def __discard_repeating_beams__(self, scan, ch=True):
        """
        Discard all more than one repeating beams

        Parameters
        ----------
        scan: Object
            Object containing a single SuperDARN scan

        Returns
        -------
        oscan: Object
            Object containing a single SuperDARN scan
            without repeating beams
        """
        oscan = Scan()
        if ch:
            scan.beams = sorted(scan.beams, key=lambda bm: (bm.bmnum))
        else:
            scan.beams = sorted(scan.beams, key=lambda bm: (bm.bmnum, bm.time))
        bmnums = []
        for bm in scan.beams:
            if bm.bmnum not in bmnums:
                if hasattr(bm, "slist") and len(getattr(bm, "slist")) > 0:
                    oscan.beams.append(bm)
                    bmnums.append(bm.bmnum)
        if len(oscan.beams) > 0:
            oscan.update_time()
        oscan.beams = sorted(oscan.beams, key=lambda bm: bm.bmnum)
        return oscan

    def __do_filter__(self, scans, params_to_run_filter=["v", "w_l",
                                                         "p_l", "elv"]):
        """
        Median filter based on the weight given by matrix (3X3X3) weight,
        and threshold based on thresh

        Parameters
        ----------
        scans: Object
            Object containing SuperDARN data in scans
        params_to_run_filter: list
            List of parameters to run boxcar filter

        Returns
        -------
        oscan: Object
            Object containing SuperDARN scans
        """
        scans = [self.__discard_repeating_beams__(s) for s in scans]
        if len(scans) == 3:
            w = self.w
            oscan = Scan()

            for b in scans[1].beams:
                bmnum = b.bmnum
                beam = Beam()
                beam.copy(b)

                for key in beam.__dict__.keys():
                    if type(getattr(beam, key)) is np.ndarray:
                        setattr(beam, key, [])

                for r in range(0, b.nrang):
                    box = [[[None for j in range(3)]
                            for k in range(3)] for n in range(3)]

                    # Iterate through time
                    for j in range(0, 3):
                        # Iterate through beam
                        for k in range(-1, 2):
                            # Iterate through gate
                            for n in range(-1, 2):
                                # Get the scan we are working on
                                s = scans[j]
                                if s is None:
                                    continue
                                # Get the beam we are working on
                                if s is None:
                                    continue
                                # Get the beam we are working on
                                tbm = None
                                for bm in s.beams:
                                    if bm.bmnum == bmnum + k:
                                        tbm = bm
                                if tbm is None:
                                    continue
                                # Check if target gate number is in the beam
                                if r + n in tbm.slist:
                                    ind = np.array(tbm.slist).tolist()\
                                            .index(r + n)
                                    box[j][k + 1][n + 1] = Gate(
                                        tbm, ind, gflg_type=self.gflg_type
                                    )
                                else:
                                    box[j][k + 1][n + 1] = 0
                    pts = 0.0
                    tot = 0.0
                    v, w_l, p_l, elv, gfx = \
                        list(), list(), list(), list(), list()

                    # Iterate through time
                    for j in range(0, 3):
                        # Iterate through beam
                        for k in range(0, 3):
                            # Iterate through gate
                            for n in range(0, 3):
                                bx = box[j][k][n]
                                if bx is None:
                                    continue
                                wt = w[j][k][n]
                                tot += wt
                                if bx != 0:
                                    pts += wt
                                    for m in range(0, wt):
                                        v.append(bx.v)
                                        w_l.append(bx.w_l)
                                        p_l.append(bx.p_l)
                                        elv.append(bx.elv)
                                        gfx = bx.gflg
                    # Check if we meet the threshold
                    if pts / tot >= self.thresh:
                        beam.slist.append(r)
                        beam.v.append(np.median(v))
                        beam.w_l.append(np.median(w_l))
                        beam.p_l.append(np.median(p_l))
                        beam.elv.append(np.median(elv))
                        beam.gflg.append(gfx)
                oscan.beams.append(beam)

            if len(oscan.beams) > 0:
                oscan.update_time()
                sorted(oscan.beams, key=lambda bm: bm.bmnum)
        else:
            oscan = Scan()
        return oscan



# Copyright 2023 SuperDARN Canada, University of Saskatchewan
# Author: Carley Martin
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# This code to calculate the solar terminator at given height in the ionosphere
# is converted from Carley Martin's JavaScript terminator.js code
#
# Modification:
# 

import numpy as np
import datetime as dt

from pydarn import Re

"""
terminator.py is a group of methods used to calculated the part of a map where
the ionosphere is in the earth's shadow
"""


def antipode(psn):
    """
    antipode calculates the opposite side of the Earth to the sub-solar 
    point

    Parameters
    ----------
    psn : list
        list containing the longitude/latitude of the sub solar point

    Returns
    -------
    antipode: lise
        list containing the logitude/latitude of the opposite side of the 
        sub-solar position
    """
    antipode = [psn[0] + 180, -psn[1]]
    return antipode


def eccentricity_earths_orbit(centuries):
    """
    eccentricity of Earths orbit calculates the ellipse created by Earth in
    it's orbit

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    eccentricity: float
        eccentricity of Earth's orbit, likely small number as all dates in
        SuperDARN operations are around 2000
    """
    # Eccentricity of Earth's orbit (unitless) = 0.016708634
    eccentricity = 0.016708634 - centuries * (0.000042037 + 0.0000001267
                                              * centuries)
    return eccentricity


def mean_obliquity_ecliptic(centuries):
    """
    mean obliquity ecliptic calculates the effect caused by the tilt of the
    Earth on its axis

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    oblic: float (radians)
        obliquity of Earths axis
    """
    obliq = np.radians(23 + (26 + (21.448 - centuries * (46.8150 + centuries *
                (0.00059 - centuries * 0.001813))) / 60) / 60)
    return obliq


def obliquity_correction(centuries):
    """
    obliquity correction corrects the obliquity of Earths axis, includes small
    perturbations in obliquity over time

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    oblic: float (radians)
        obliquity of Earths axis
    """
    mean_ob = mean_obliquity_ecliptic(centuries)
    obliq = mean_ob + np.radians(0.00256) * np.cos(np.radians(125.04 - 1934.136
                                                              * centuries))
    return obliq


def solar_geometric_mean_anomaly(centuries):
    """
    solar geometric mean anomaly calculates the apparent motion of the sun and
    along the plane of the ecliptic

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    anomaly: float (radians)
        angular distance from perihelion which Earth would have moved around
        Sun if using constant angular velocity
    """
    anomaly = np.radians(357.52911 + centuries
                         * (35999.05029 - 0.0001537 * centuries))
    return anomaly


def solar_geometric_mean_longitude(centuries):
    """
    solar geometric mean longitude calculates where the mean Sun was at given
    date time in reference to Julian base date

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    mean_long: float (radians)
        solar meal longitude
    """
    # Mean longitude of sun at base julian date = 280.46646 degrees
    # Corrections from J.Meeus Astronomical Algorithms book
    el = (280.46646 + centuries * (36000.76983 + centuries * 0.0003032)) % 360
    # Give value between 0 and 360
    if el < 0:
        el = el + 360
    mean_long = np.radians(el)
    return mean_long


def solar_equation_center(centuries):
    """
    solar equation center calculates the angular difference between the
    actual position of a body and it's position if it's motion was uniform in
    a circular orbit - approx from Keppler's equation allowed as the
    eccentricity is very small

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    solar_center: float (radians)
        solar center angle
    """
    m = solar_geometric_mean_anomaly(centuries)
    solar_center = np.radians(np.sin(m)
                   * (1.914602 - centuries * (0.004817 + 0.000014 * centuries))
                   + np.sin(m + m) * (0.019993 - 0.000101 * centuries)
                   + np.sin(m + m + m) * 0.000289)
    return solar_center


def solar_true_longitude(centuries):
    """
    solar true longitude calculates the true longitude of the Sun

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    true_long: float (radians)
        the true longitude
    """
    true_long = solar_geometric_mean_longitude(centuries)\
                + solar_equation_center(centuries)
    return true_long


def solar_apparent_longitude(centuries):
    """
    solar apparent longitude calculates the true longitude including
    extra perturbations from the Moon etc. 

    Parameters
    ----------
    centuries: float
        decimal value of centuries passed since Julian base

    Returns
    -------
    apparent: float (radians)
        the true apparent longitude
    """
    apparent = solar_true_longitude(centuries) - np.radians(0.00569 + 0.00478
                    * np.sin(np.radians(125.04 - 1934.136 * centuries)))
    return apparent


def solar_declination(centuries):
    dec = np.arcsin( np.sin(obliquity_correction(centuries))\
          * np.sin(solar_apparent_longitude(centuries)))
    return dec


def equation_of_time(centuries):
    # Equation from NOAA subsolar calculator
    # http://www.esrl.noaa.gov/gmd/grad/solcalc/
    e = eccentricity_earths_orbit(centuries)
    m = solar_geometric_mean_anomaly(centuries)
    el = solar_geometric_mean_longitude(centuries)
    y = np.tan(obliquity_correction(centuries) /2)
    time_eq = y * np.sin(2 * el) - 2 * e * np.sin(m) + 4 * e * y * np.sin(m)\
              * np.cos(2 * el) - 0.5 * y * y * np.sin(4 * el)\
              - 1.25 * e * e * np.sin(2 * m)
    return time_eq


def solar_position(date):
    """
    solar position calculates the latitude and longitude of the sub-solar
    position on Earth for a given date time

    Parameters
    ----------
    date : datetime object
            date time of interest

    Returns
    -------
    psn: list
        list containing the logitude/latitude of the 
        sub-solar position
    """
    # Convert Greg date to Julian date (given in partial centuries)
    # 86400 seconds = 1 days
    # 36525 days = 100 years (a century)
    centuries = (date - dt.datetime(2000,1,1,12,0)).total_seconds() / 86400 / 36525
    # Given days date but at midnight
    datefloor = date.replace(hour=0, minute=0, second=0)
    # Calculate the longitude at given time
    longitude = (datefloor - date).total_seconds() / 86400 * 360 - 180
    # Calculate long lat of sub-solar position in degrees
    long = longitude - np.degrees(equation_of_time(centuries))
    lat = np.degrees(solar_declination(centuries))
    return [long, lat]


def antisolar(date):
    """
    antisolar calculates the latitude and longitude of the anti-sub-solar
    position on Earth for a given date time

    Parameters
    ----------
    date : datetime object
            date time of interest

    Returns
    -------
    antisolar_point: list
        list containing the logitude/latitude of the 
        anti-sub-solar position
    """
    antisolar_point = antipode(solar_position(date))
    return antisolar_point


def terminator(date, height):
    # Get the anti-sub-solar point
    antisolar_point = antisolar(date)
    # Calculate the size of the great circle (radius from anti-sub-solar point)
    arc_angle = 90 - np.degrees(np.arccos(Re / (Re + height)))
    # Calculate arc in km
    arc_length = Re * np.radians(arc_angle)
    
    return antisolar_point, arc_length, arc_angle




# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
# This code is based on acf.py in the DaVitpy library
# https://github.com/vtsuperdarn/davitpy/blob/master/davitpy
#
# Modifications:
# 2022-05-03: CJM - Added options to plot power and phase of acf
#                 - change defaults to fit needs
# 2023-06-28: CJM - Refactored return values
# 2023-12-15: RAR - Made helper function for plotting ACF values
#                 - Refactored plot_acfs() to use helper function for plotting
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.


import matplotlib.markers
import matplotlib.pyplot as plt
import numpy as np
import warnings
import matplotlib.ticker as ticker

from datetime import datetime
from typing import List

from pydarn import (plot_exceptions, SuperDARNRadars,
                    standard_warning_format, time2datetime,
                    check_data_type, RadarID)

warnings.formatwarning = standard_warning_format


class ACF:
    """
    ACF class plots SuperDARN RAWACF data

    Class pattern design: Builder
    This class inherits matplotlib.pyplot to inherit plotting features as well
    build off their builder design pattern.

    Notes
    -----
    This class extends from matplotlib to generate the figures

    Methods
    -------
    plot
    plot_amplitude (not implemented yet)
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_acfs()\n"\


    @staticmethod
    def plot_acfs(dmap_data: List[dict], beam_num: int = 0,
                  gate_num: int = 15, parameter: str = 'acfd',
                  scan_num: int = 0, start_time: datetime = None,
                  ax: matplotlib.axes.Axes = None,
                  normalized: bool = False, real_color: str = 'red',
                  plot_blank: bool = True, blank_marker: str = 'o',
                  imaginary_color: str = 'blue', legend: bool = True,
                  pwr_and_phs: bool = True, pwr_color: str = 'tab:orange',
                  phs_color: str = 'tab:purple',
                  real_marker: str = 'o', imaginary_marker: str = 'o',
                  pwr_marker: str = 'o', phs_marker: str = 'o',
                  **kwargs):
        """
        Plots the parameter ACF/XCF field from SuperDARN file,
        typically RAWACF format for a given beam and gate number

        Parameters
        ----------
        dmap_data : list[dict]
            records from a dmap file
        beam_num : int
            the beam number to plot, default: 0
        gate_num : int
            the gate number to plot, default: 15
        parameter : str
            the parameter to plot, default: acfd
        scan_num: int
            will plot the scan number for the provided beam
        start_time: datetime object
            datetime object containing the record to plot for
            the given beam
        ax: matplotlib.axes.Axes
            axes object for another way of plotting. **WILL BE OVERWRITTEN IF PLOTTING POWER AND PHASE**
            Default: None
        normalized: bool
            normalizes the parameter data with the associated
            power 0 value for the given gate number
            default: False
        real_color: str
            line color of the real part of the parameter
            default: red
        plot_blank: bool
            boolean to determine if blanked lags should be plotted
            default: False
        blank_marker: str
            the marker symbol of blanked lags
            default: o - dot
        imaginary_color: str
            line color of the imaginary part of the parameter
            default: blue
        legend: bool
            produces a standard legend
            default: True
        pwr_and_phs: bool
            plots the power and phase of the ACF
            default: True
        pwr_color: str
            the color of line and symbol of power
            default: tab:orange
        phs_color: str
            the color of line and symbol of phase
            default: tab:purple
        real_marker: str
            the marker symbol of real lags
            default: o - dot
        imaginary_marker: str
            the marker symbol of imaginary lags
            default: o - dot
        pwr_marker: str
            the marker symbol of power
            default: o - dot
        phs_marker: str
            the marker symbol of phase
            default: o - dot
        kwargs: dict
            are applied to the plots

        Notes
        -----
        plot blanked lags as hollow circles

        Raises
        ------
        UnknownParameterError
        IncorrectPlotMethodError
        OutOfRangeGateError

        Returns
        -------
        dict of standard parameters, including:
            ax: list of axes that were plotted on
            fig: Figure that contains plot
            ccrs: None,
            cm: None,
            cb: None,
            data: dict of parameters plotted.
                real: real component of ACF,
                imaginary: imaginary component of ACF,
                power: Power of ACF,
                phase: Phase of ACF,
                blanked: Blanked lags of the ACF
        """
        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            index_first_match = next(i for i, d in enumerate(dmap_data)
                                     if parameter in d)
        except StopIteration:
            raise plot_exceptions.UnknownParameterError(parameter)
        check_data_type(dmap_data, parameter, 'array', index_first_match)

        # search over the records to find the correct beam and scan/time
        # to plot the corresponding ACF/XCF plot
        scan_count = 0
        re = []
        im = []
        for record in dmap_data:
            if record['bmnum'] == beam_num:
                time = time2datetime(record)
                if start_time is not None:
                    if time.day != start_time.day or \
                       time.month != start_time.month or \
                       time.year != start_time.year:
                        raise plot_exceptions.IncorrectDateError(time,
                                                                 start_time)

                if ACF.__found_scan(scan_num, scan_count, start_time, time):
                    if gate_num >= record['nrang'] or gate_num < 0:
                        raise plot_exceptions.\
                                OutOfRangeGateError(parameter, gate_num,
                                                    record['nrang'])
                    # get the difference to get the lag number
                    lags = [lag[1] - lag[0] for lag in record['ltab']]
                    # remove the last lag as it is the zeroth lag
                    lags = lags[:-1]

                    # grab corresponding real and imaginary parts
                    # of the ACF/XCF field for a given gate number
                    re = [x[0] for x in record[parameter][gate_num]]
                    im = [x[1] for x in record[parameter][gate_num]]

                    # calculate the blank lags for this record and gate
                    blanked_lags = ACF.__blanked_lags(record, lags, gate_num)
                    lag_num = 0
                    lag_idx = 0

                    # Create a mask for hiding missing or blanked lags
                    mask = [0] * len(lags)

                    # Search for missing lags
                    # Note: had to use while loop do to insert method
                    while lag_idx < len(lags):
                        if lag_num in blanked_lags:
                            mask[lag_idx] = 1   # This lag is blanked, mask it
                        if lags[lag_idx] != lag_num:
                            lags.insert(lag_num, lag_num)
                            mask.insert(lag_num, 1) # this lag is missing, mask it
                            # Adding in nan's for missing lags (won't be plotted)
                            re.insert(lag_num, np.nan)
                            im.insert(lag_num, np.nan)

                        lag_idx += 1
                        if lag_idx < len(lags):
                            if lag_num != lags[lag_idx]:
                                if lag_num > lags[lag_idx]:
                                    lag_num = lags[lag_idx]
                                else:
                                    lag_num += 1
                    # once we find the data break free!!
                    break
                scan_count += 1
        if record['cp'] == 503:
            warnings.warn("Please note that this data is from Tauscan "
                          "which has different lag properties compared "
                          "to other control programs. The ACF plot may "
                          "not be correct. Please contact the PI of the "
                          "radar to confirm if the data looks correct.")
        if re == [] or im == []:
            if 0 < gate_num < record['nrang']:
                raise plot_exceptions.\
                    NoDataFoundError(parameter, beam_num,
                                     opt_beam_num=record['bmnum'])
            else:
                raise plot_exceptions.OutOfRangeGateError(parameter, gate_num,
                                                          record['nrang'])

        if normalized is True:
            re /= record['pwr0'][gate_num]
            im /= record['pwr0'][gate_num]

        # generates gaps where there are missing or blanked lags
        masked_re = np.ma.array(re, mask=mask)
        masked_im = np.ma.array(im, mask=mask)

        # Calculate pwr and phs regardless of choice as they are returned
        pwr = np.sqrt(np.square(re) + np.square(im))
        phs = np.rad2deg(np.arctan2(im, re))
        masked_pwr = np.ma.array(pwr, mask=mask)
        masked_phs = np.ma.array(phs, mask=mask)

        # plot real and imaginary with power
        if pwr_and_phs is True:
            fig = plt.gcf()
            gs = fig.add_gridspec(2, 2)
            ax = fig.add_subplot(gs[0, :])
            ax_pwr = fig.add_subplot(gs[1, 0])
            ax_phs = fig.add_subplot(gs[1, 1])
            axes = [ax, ax_pwr, ax_phs]
        else:
            if ax is not None:
                ax = ax
            else:
                ax = plt.gca()
            axes = [ax]

        real_dict = ACF.plot_acf_param(lags, masked_re, ax,
                                       plot_blank=plot_blank,
                                       marker=real_marker,
                                       color=real_color,
                                       blank_marker=blank_marker,
                                       **kwargs)
        imag_dict = ACF.plot_acf_param(lags, masked_im, ax,
                                       plot_blank=plot_blank,
                                       marker=imaginary_marker,
                                       color=imaginary_color,
                                       blank_marker=blank_marker,
                                       **kwargs)
        # Set parameters (axis labels, line labels, etc.) of main plot
        ax.set_ylabel(parameter)
        ax.set_xlabel('Lag Number')
        real_dict['data']['line'].set_label('Real')
        imag_dict['data']['line'].set_label('Imag')
        if plot_blank:
            real_dict['data']['blank_line'].set_label('Real Blanked')
            imag_dict['data']['blank_line'].set_label('Imag Blanked')
            lim_val = 1.1 * max(np.nanmax(np.abs(re)), np.nanmax(np.abs(im)))
        else:
            lim_val = 1.1 * max(np.abs(masked_re).max(),
                                np.abs(masked_im).max())
        ax.set_ylim([-lim_val, lim_val])
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))

        # Plot the power and phase, if specified
        if pwr_and_phs is True:
            power_dict = ACF.plot_acf_param(lags, masked_pwr, ax_pwr,
                                            plot_blank=plot_blank,
                                            marker=pwr_marker,
                                            color=pwr_color,
                                            blank_marker=blank_marker,
                                            **kwargs)
            ax_pwr.set_ylabel('Power')
            ax_pwr.set_xlabel('Lag Number')
            ax_pwr.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
            ax_pwr.set_ylim([0, 1.1 * np.nanmax(pwr)])
            ax_pwr.set_title('Power')

            phase_dict = ACF.plot_acf_param(lags, masked_phs, ax_phs,
                                            plot_blank=plot_blank,
                                            marker=phs_marker,
                                            color=phs_color,
                                            blank_marker=blank_marker,
                                            **kwargs)
            ax_phs.set_ylabel('Phase (degrees)')
            ax_phs.set_xlabel('Lag Number')
            ax_phs.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
            ax_phs.set_ylim([-180, 180])
            ax_phs.set_title('Phase')

            # Set labels on the plots
            power_dict['data']['line'].set_label('Power')
            phase_dict['data']['line'].set_label('Phase')
            if plot_blank:
                power_dict['data']['blank_line'].set_label('Power Blanked')
                phase_dict['data']['blank_line'].set_label('Phase Blanked')

        # Make legend on right side of figure
        if legend is True:
            fig = plt.gcf()
            fig.legend(loc=5)
            plt.subplots_adjust(right=0.8)

        # Set title of Plot
        radar_name = SuperDARNRadars.radars[RadarID(dmap_data[0]['stid'])].name
        title = "{date} UT {radar} Beam {beam}, Gate {gate}, Control "\
                "Program: {cpid}"\
                "".format(radar=radar_name, beam=beam_num, gate=gate_num,
                          date=time.strftime("%Y %b %d %H:%M"),
                          cpid=record['cp'])
        ax.set_title(title)

        return {'ax': axes,
                'ccrs': None,
                'cm': None,
                'cb': None,
                'fig':  plt.gcf(),
                'data': {'real': masked_re,
                         'imaginary': masked_im,
                         'power': masked_pwr,
                         'phase':  masked_phs,
                         'blanked': blanked_lags}
                }

    @staticmethod
    def plot_acf_param(x, y, ax,
                       plot_blank: bool = True,
                       blank_marker: str = 'o',
                       **kwargs):
        """
        plots the parameter ACF/XCF field from SuperDARN file,
        typically RAWACF format for a given beam and gate number

        Parameters
        ----------
        x: np.array
            X coordinates of values to plot
        y: np.ma.array
            Y coordinates of values to plot
        ax: matplotlib.Axes
            Axes object for another way of plotting
            Default: None
        plot_blank: bool
            boolean to determine if blanked lags should be plotted
            default: False
        blank_marker: str
            the marker symbol of blanked lags
            default: o - circle
        kwargs: dict
            keyword arguments for setting plot features


        Notes
        -----
        plot blanked lags as hollow circles

        Returns
        -------

        """
        line = ax.plot(x, y, **kwargs)

        # Plot blanked lags
        if plot_blank:
            mask = y.mask

            # Invert the mask so the blanked data is plotted
            blank_mask = ~mask
            y.mask = blank_mask

            # Use blank_marker as marker
            kwargs['marker'] = blank_marker

            # Check if blank_marker is a filled marker like circle or square, and set face white if so
            if blank_marker in matplotlib.markers.MarkerStyle.filled_markers:
                kwargs['facecolors'] = (1, 1, 1, 0)     # Make the marker face white
                kwargs['edgecolors'] = kwargs.get('color', None)
            # For unfilled markers like 'x', just set the color
            else:
                kwargs['color'] = kwargs.get('color', None)

            # Plot the blanks as scatter points
            blank_scatter = ax.scatter(x, y, **kwargs)

            # Now put the mask back!
            y.mask = mask
        else:
            blank_scatter = None

        return {'ax': ax,
                'ccrs': None,
                'cm': None,
                'cb': None,
                'fig': plt.gcf(),
                'data': {'x': x,
                         'y': y,
                         'line': line[0],
                         'blank_line': blank_scatter}
                }

    @staticmethod
    def __found_scan(scan_num: int, count_num: int,
                     start_time: datetime, time: datetime):
        """
        Method to help do the complicated if statement
        for scan/time check

        Parameter
        ---------
            scan_num: int
                scan number the user passed in
            count_num: int
                the current scan count
            start_time: datetime
                a datetime object the user passes in
            time: datetime
                current time the record is at

        Return:
        ------
            Bool
            If the scan is found return True
            else False
        """
        if start_time is None:
            if scan_num == count_num:
                return True
        elif start_time < time:
            return True

        return False

    @staticmethod
    def __blanked_lags(record: dict, lags: list, gate: int):
        """
        Determines the blanked lags in the data record
        Lags contaminated by transmit pulse overlap

        Parameters
        ----------
        record : dict
            data record containing the lag and ACF/SCF data
        lags : list
            the lag number list where elments are: ltab[i][1] - ltab[i][0]
        gate : int
            gate number
        """
        sorted_lags = lags
        lags_pair = [lag for lags in record['ltab'] for lag in lags]
        sorted_lags.sort()
        lags_pair.sort()
        tau_per_txpl = record['mpinc'] / record['txpl']
        blanked_samples = [lag_pair * tau_per_txpl for lag_pair in lags_pair]
        blanked_lags = []
        for blank in blanked_samples:
            blanked_lags.extend([blank, blank + 1])

        blanked_lag_numbers = []
        for s_lag in sorted_lags:
            for pair in record['ltab']:
                # find which lags didn't get changed
                # in the sorting
                if s_lag == (pair[1] - pair[0]):
                    sample1 = tau_per_txpl * pair[0] + gate + 1 * \
                        record['lagfr'] / record['txpl']
                    sample2 = tau_per_txpl * pair[1] + gate + 1 * \
                        record['lagfr'] / record['txpl']
                    if sample1 in blanked_lags or sample2 in blanked_lags:
                        blanked_lag_numbers.append(s_lag)

        return blanked_lag_numbers



import matplotlib
from matplotlib import colormaps as cm


class PyDARNColormaps():
    PYDARN_INFERNO = cm.get_cmap('inferno')
    PYDARN_PLASMA = cm.get_cmap('plasma')
    PYDARN_PLASMA_R = cm.get_cmap('plasma_r')
    PYDARN_VELOCITY = matplotlib.colors.LinearSegmentedColormap.from_list(
                        "pydarn_velocity", ["darkred", "r", 'pink', "b",
                                            "darkblue"])

    PYDARN_VIRIDIS = matplotlib.colors.LinearSegmentedColormap.from_list(
                        "pydarn_viridis", ["indigo", "midnightblue",
                                           "navy", "mediumblue",
                                           "teal", "mediumseagreen",
                                           "limegreen", "yellowgreen",
                                           "yellow", "gold"])

    PYDARN = matplotlib.colors.LinearSegmentedColormap.from_list("pydarn",
                ["midnightblue", "darkblue", "mediumblue", "rebeccapurple",
                 "purple", "darkmagenta", "mediumvioletred", "crimson", "red",
                 "orangered", "darkorange", "orange", "gold"])




# (C) Copyright 2021 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
#

"""
This file contains the version number of pydarn
"""

__version__='4.1.1'



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
# Modifications:
# 20230623 - CJM - Removed checks for read_dmap, will read in any dmap

import bz2
import pydarnio
import os


class SuperDARNRead(pydarnio.SDarnRead):
    """
    A class that reads select SuperDARN files for pyDARN plotting

    Methods
    -------
    read_dmap : reads superDARN DMap formats
    read_borealis: Reads Borealis hdf5 formats and converts
        Borealis' data dictionary to SDARN data dictionary
    """
    def __init__(self, filename: str = None, stream: bool = False):
        if filename is not None:
            super().__init__(filename, stream)

    def read_dmap(self, filename: str):
        """
        Reads select SuperDARN DMap files for pyDARN plotting:
            fitacf
            rawacf
            iqdat
        When more plotting methods are introduced other files
        will be included.

        Parameters
        ----------
            filename: string
                name of the file you are going to read

        Returns
        ------
            data: List[dict]
                data records from the file
        """
        # check if the file  is compressed with
        # bz2
        if 'bz2' in filename:
            with bz2.open(filename) as fp:
                dmap_stream = fp.read()

            super().__init__(dmap_stream, stream=True)
        else:
            super().__init__(filename)

        try:
            # Check which file type it is
            data = self.read_records()
        except Exception as err:
            # sometimes files might fail due to issues with
            # the specific reading methods. This is not a pyDARN
            # issue so we will just read the dmap file if that fails
            # then make an issue on pyDARNio
            print(err)
            print("..... Will try to read DMap file with read_dmap")
            print(" IF THIS FAILS please make an issue on the pyDARNio"
                  " GitHub repository:"
                  " https://github.com/SuperDARN/pyDARNio/issues")
            data = self.read_records
        return data

    def read_borealis(self, filename: str, slice_id: int = None):
        """
        Reads RAWACF or BFIQ borealis files and converts them to
        an SDARN data format dictionary for plotting.

        Parameters
        ----------
            filename: str
                string name of the file, make sure "rawacf" or "bfiq" is in
                the name
            slice_id: int
                the Borealis slice id of the file, required if reading Borealis
                data produced prior to when Borealis v0.5 was released

        Post
        ----
            Produces a file named: dmap_file.rawacf, this should be deleted
            at the end of the file. Requirements for BorealisConvert.

        Raises
        -----
            ValueError - if the file type is not determined in the filename
            then it raises an error that it cannot convert the file.
        """
        new_filename = 'dmap_file.rawacf'

        if 'rawacf' in filename:
            file_type = 'rawacf'
        elif 'bfiq' in filename:
            file_type = 'bfiq'
        else:
            raise ValueError("The filetype of {} cannot  be determined,"
                             " please make sure the filename has rawacf"
                             " or bfiq in the name or look up the accepted"
                             " file in pyDARNio's"
                             " documentation".format(filename))

        converter = pydarnio.BorealisConvert(filename, file_type,
                                             new_filename, slice_id)
        os.remove(new_filename)
        return converter.sdarn_dict



# Copyright (C) SuperDARN Canada, Universtiy of Saskatchewan
# Authors: Marina Schmidt
import logging

pydarn_log = logging.getLogger('pydarn')


class RTPZeroError(Exception):
    """
    Error raised when a ZeroDivisionError is raised due to minimum - maximum
    range cause mathematical errors in the plotting. An example of this is
    when pwr0 is plotted, some files will have a pwr0=0 which will cause
    a ZeroDivisionError using logNorm as the normalization method. This
    cannot be avoided and challenging to detect which normalization method
    is passed in. Thus error will be raised to let the user know to change
    the range or normalization method to fix the problem.
    """
    def __init__(self, parameter: str, beam_num: int, zmin: float, zmax: float,
                 norm):
        self.parameter = parameter
        self.beam_num = beam_num
        self.zmin = zmin
        self.zmax = zmax
        self.norm = norm
        self.message = "ZeroDivisionError or ValueError is raised "\
            "due to trying to normalize {zmin} to {zmax} using the {norm} "\
            "method for the parameter {param} at beam {num}. "\
            "This can be an issue with the data file, fields meaning"\
            " something different. To resolve this error, please set the "\
            "zmin or zmax parameter or try another normalization"\
            " method.".format(zmin=zmin, zmax=zmax, norm=norm,
                              num=beam_num, param=parameter)
        super().__init__(self.message)
        pydarn_log.error(self.message)



# (C) Copyright 2021 SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
# 2022-03-10 MTS added 4 new methods to generate coordinates for the various
#                enums
# 2023-08-26 CJM corrected calculations to use bmoff and removed abs()
#

"""
coordinates.py is a group of methods that focus on coordinates systems used in
plotting
"""
import datetime as dt
import enum
import numpy as np

import aacgmv2

import pydarn
from pydarn import (geocentric_coordinates, SuperDARNRadars, RangeEstimation,
                    radar_exceptions, Re, RadarID)


def geo_coordinates(stid: RadarID, beams: int = None,
                    gates: tuple = None, **kwargs):
    """
    geographic_coordinates calculates the geographic coordinate for a given
    set of gates and beams
    parameters
    -----------

    """
    if gates is None:
        gates = [0, SuperDARNRadars.radars[stid].range_gate_45]
    if beams is None:
        beams = SuperDARNRadars.radars[stid].hardware_info.beams

    # Plus 1 is due to the fact fov files index at 1 so in the plotting
    # of the boundary there is a subtraction of 1 to offset this as python
    # converts to index of 0 which my code already accounts for
    beam_corners_lats = np.zeros((gates[1]-gates[0]+1, beams+1))
    beam_corners_lons = np.zeros((gates[1]-gates[0]+1, beams+1))

    for beam in range(0, beams+1):
        for gate in range(gates[0], gates[1]+1):
            lat, lon = gate2geographic_location(stid=stid, beam=beam,
                                                range_gate=gate, height=300,
                                                **kwargs)
            beam_corners_lats[gate-gates[0], beam] = lat
            beam_corners_lons[gate-gates[0], beam] = lon
    y0inx = np.min(np.where(np.isfinite(beam_corners_lats[:,0]))[0])
    return beam_corners_lats[y0inx:], beam_corners_lons[y0inx:]


def aacgm_coordinates(stid: pydarn.RadarID, beams: int = None, gates: tuple = None,
                      date: dt.datetime = dt.datetime.now, **kwargs):
    if gates is None:
        gates = [0, SuperDARNRadars.radars[stid].range_gate_45]
    if beams is None:
        beams = SuperDARNRadars.radars[stid].hardware_info.beams

    # Plus 1 is due to the fact fov files index at 1 so in the plotting
    # of the boundary there is a subtraction of 1 to offset this as python
    # converts to index of 0 which my code already accounts for
    beam_corners_lats = np.zeros((gates[1]-gates[0]+1, beams+1))
    beam_corners_lons = np.zeros((gates[1]-gates[0]+1, beams+1))

    for beam in range(0, beams+1):
        for gate in range(gates[0], gates[1]+1):
            lat, lon = gate2geographic_location(stid=stid, beam=beam,
                                                range_gate=gate, height=300,
                                                **kwargs)
            geomag = np.array(aacgmv2.get_aacgm_coord(glat=lat,
                                                      glon=lon,
                                                      height=250,
                                                      dtime=date))
            beam_corners_lats[gate-gates[0], beam] = geomag[0]
            beam_corners_lons[gate-gates[0], beam] = geomag[1]
    y0inx = np.min(np.where(np.isfinite(beam_corners_lats[:,0]))[0])
    return beam_corners_lats[y0inx:], beam_corners_lons[y0inx:]


def aacgm_MLT_coordinates(**kwargs):
    beam_corners_lats, beam_corners_lons = aacgm_coordinates(**kwargs)
    beam_corners_mlts = convert2MLT(beam_corners_lons, **kwargs)
    return beam_corners_lats, beam_corners_mlts


def convert2MLT(lons: float, date: object, **kwargs):
    fan_shape = lons.shape
    # Work out shift due in MLT
    beam_corners_mlts = np.zeros((fan_shape[0], fan_shape[1]))
    mltshift = lons[0, 0] - (aacgmv2.convert_mlt(lons[0, 0], date) * 15)
    beam_corners_mlts = lons - mltshift
    return beam_corners_mlts


def gate2geographic_location(stid: pydarn.RadarID, beam: int, height: float = None,
                             elv_angle: float = 0.0, center: bool = False,
                             range_estimation: RangeEstimation =
                             RangeEstimation.SLANT_RANGE, **kwargs):
    """
    determines the geographic cell position for a given range gate and beam
    Notes: From RPosGeo line 335

    parameters
    ----------
        stid: pydarn.RadarID
            station id of the radar to use
        beam: int
            beam number (indexing at 0)
        height: float
            transmutation height [km]
            default: none
            if none then it uses elevation angle
        elv_angle: float
            elevation angle in [deg]
            default: 0
        center: bool
            obtain geographic location of the centre of the range gates
            False obtains the near-left corner of the range gates.
            See also: gate2slant in range_estimation module
            default: False (return corner values)

    returns
    -------
        lat: float
            latitude of the range gate in geographic coordinates [deg]
        lon: float
            longitude of the range gate in geographic coordinates [deg]
    """
    # centre of the field of view
    offset = SuperDARNRadars.radars[stid].hardware_info.beams / 2.0 - 0.5

    # Obtain radar information from hardware files all converted to [rad]
    # radians are needed for numpy geometry calculations and to reduce
    # too much converting back and forth.
    boresight = np.radians(SuperDARNRadars.
                           radars[stid].hardware_info.boresight.physical)
    bmoff = np.radians(SuperDARNRadars.
                       radars[stid].hardware_info.boresight.electronic)
    radar_lat = np.radians(SuperDARNRadars.
                           radars[stid].hardware_info.geographic.lat)
    radar_lon = np.radians(SuperDARNRadars.
                           radars[stid].hardware_info.geographic.lon)
    # Note that some beam separations are negative
    beam_sep = np.radians(SuperDARNRadars.
                          radars[stid].hardware_info.beam_separation)
    rxrise = SuperDARNRadars.radars[stid].hardware_info.rx_rise_time

    # If the user wants the edge corner of the range gate:
    # Radar outwards direction center value is corrected in
    # range_estimation module
    if center:
        beam_edge = 0
    else:
        beam_edge = -beam_sep * 0.5

    # psi [rad] in the angle from the boresight
    psi = beam_sep * (beam - offset) + beam_edge + bmoff
    # Calculate the slant range [km]
    if range_estimation == RangeEstimation.RANGE_GATE:
        raise radar_exceptions.RangeEstimationError("Range gates cannot be "
                                                    "used in to estimate "
                                                    "distance. Try SLANT_RANGE"
                                                    " instead.")
    elif range_estimation == RangeEstimation.TIME_OF_FLIGHT:
        raise radar_exceptions.RangeEstimationError("Time of flight cannot be "
                                                    "used in to estimate "
                                                    "distance. Try SLANT_RANGE"
                                                    " instead.")
    else:
        target_range = range_estimation(rxrise=rxrise, **kwargs)

    # If no height is specified then use elevation angle (default 0)
    # to calculate the transmutation height
    if height is None:
        height = -Re + np.sqrt(Re**2 + 2 * target_range * Re *
                               np.sin(np.radians(elv_angle)) + target_range**2)

    lat, lon = geocentric_coordinates(lat=radar_lat,
                                      lon=radar_lon,
                                      target_range=target_range,
                                      cell_height=height,
                                      psi=psi,
                                      boresight=boresight, **kwargs)
    # convert back degrees as preferred units to use?
    return np.degrees(lat), np.degrees(lon)


class Coords(enum.Enum):
    """
    This coordinate class is to list the current coordinate systems
    a user can pick from

    enumerators:
        RANGE_GATE: range gates
        SLANT_RANGE: slant range (km)
        GROUND_SCATTER_MAPPED_RANGE: ground scatter mapped range (km)
        GEOGRAPHIC: geographical coordinates lat and longitude (degree)
        AACGM: Magnetic geographical coordinates lat and longitude (degree)
    """

    GEOGRAPHIC = (geo_coordinates, )
    AACGM = (aacgm_coordinates, )
    AACGM_MLT = (aacgm_MLT_coordinates, )

    # Need this to make the functions callable
    def __call__(self, *args, **kwargs):
        return self.value[0](*args, **kwargs)



# Copyright (C) 2020 SuperDARN Canada, Universtiy of Saskatchewan
# Author(s): Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modification:
"""
This module is utility functions that are useful
for multiple plotting methods
"""
import datetime as dt
import enum
import matplotlib.pyplot as plt
import numpy as np
import warnings

from typing import List

from pydarn import plot_exceptions

class MapParams(enum.Enum):
    """
    """
    FITTED_VELOCITY = "fitted"
    MODEL_VELOCITY = "model.vel.median"
    RAW_VELOCITY = "vector.vel.median"
    POWER = "vector.pwr.median"
    SPECTRAL_WIDTH = "vector.wdt.median"

class TimeSeriesParams(enum.Enum):
        """
        Enum class to hold the parameters that can be plotted
        Members
        ------------
        NUM_VECTORS:
        the number of aviliable vectors in the map
        IMF_BY:
        The magnitude of the IMF By component in nT
        IMF_BZ:
        The magnitude of the IMF Bz component in nT
        KP:
        The Kp index
        HMB_MAX:
        The farthest equatorward extent of the HMB

        """
        NUM_VECTORS = 'num_vectors'
        IMF_BY = 'IMF.By'
        IMF_BZ = 'IMF.Bz'
        IMF_BX = 'IMF.Bx'
        IMF_Vx = 'IMF.Vx'
        IMF_TILT = 'IMF.tilt'
        KP = 'IMF.Kp'
        LATMINN = 'latmin'
        ERR = 'chi.sqr.dat'
        CPP = 'pot.drop'
        POT = 'pot.pos'
        

def find_record(dmap_data: List[dict], start_time: dt.datetime, time_delta:int = 1):
    """
    finds the record number that associates to the start time

    Parameter
    ---------
        dmap_data : List[dict]
            the data to look over for the record number
        start_time : datetime
            the start_time to associate to the record number
        time_delta : int
            the difference between start_time and dmap_data time to determine
            the record number within a region

    Return
    ------
        record_num : int
            the record number associated with the start_time

    Raises
    ------
        NoDataFound
            raises if the start_time is not in the dmap_data list
    """
    record_num = 0
    for record in dmap_data:
        date = time2datetime(record)
        time_diff = date - start_time
        if time_diff.seconds/60 <= time_delta:
            return record_num
        record_num += 1
    raise plot_exceptions.NoDataFoundError('N/A', start_time=start_time)

def check_data_type(dmap_data: List[dict], parameter: str,
                    expected_type: str, index: int):
    """
    Checks to make sure the plot type is correct
    for the data structure

    Parameters
    ----------
    dmap_data: List(dict)
        list of records (dictionaries) representing
        DNap data
    parameter: str
        string key word name of the parameter
    expected_type: str
        string describing an array or scalar type
        to determine which one is needed for the type of plot
    index: int
        record number to check
    Raises
    -------
    RTPIncorrectPlotMethodError
    """
    data_type = dmap_data[index][parameter]
    if expected_type == 'array':
        if not isinstance(data_type, np.ndarray):
            raise plot_exceptions.IncorrectPlotMethodError(parameter,
                                                           data_type)
    else:
        if isinstance(data_type, np.ndarray):
            # TODO: make into a general plotting exception
            raise plot_exceptions.IncorrectPlotMethodError(parameter,
                                                           data_type)


def time2datetime(dmap_record: dict) -> dt.datetime:
    """
    Converts DMAP time parameter fields into a datetime object

    Parameter
    ---------
    dmap_record: dict
        dictionary of the DMAP data contains the time data

    Returns
    -------
    datetime object
        returns a datetime object of the records time stamp
    """
    try:
        year = dmap_record['time.yr']
        month = dmap_record['time.mo']
        day = dmap_record['time.dy']
        hour = dmap_record['time.hr']
        minute = dmap_record['time.mt']
        second = dmap_record['time.sc']
        micro_sec = dmap_record['time.us']
        return dt.datetime(year=year, month=month, day=day, hour=hour,
                           minute=minute, second=second, microsecond=micro_sec)
    except KeyError:
        year = dmap_record['start.year']
        month = dmap_record['start.month']
        day = dmap_record['start.day']
        hour = dmap_record['start.hour']
        minute = dmap_record['start.minute']
        second = dmap_record['start.second']
        return dt.datetime(year=year, month=month, day=day, hour=hour,
                           minute=minute, second=int(second))


def add_embargo(fig: plt.Figure):
    """
    Adds a watermark to the figure noting that the data is under
    embargo

    Parameter
    ---------
    fig: plt.Figure
    """
    vals = []
    for t in range(0, len(fig.texts)):
        vals.append(fig.texts[t].get_text())
    if not any(item == 'EMBARGOED' for item in vals):
        fig.text(0.5, 0.5, "EMBARGOED", fontsize=70,
                 color='grey', ha='center', va='center',
                 rotation=-20, alpha=0.7)


def determine_embargo(time: dt.datetime, cpid: int, radar: str):
    """
    Determines if the data is under embargo

    Parameter
    ---------
    time: dt.datetime
        Timestamp of the data
    cpid: int
        Control program ID
    radar: str
        Name of the radar which produced the data

    Returns
    -------
    bool: True if data is embargoed, False otherwise
    """
    year_ago = dt.datetime.now() - dt.timedelta(days=365)
    if time > year_ago and cpid < 0:
        warnings.warn(f'The data you are using is under embargo. '
                      f'Please contact the principal investigator '
                      f'of the {radar} radar for authorization to '
                      f'use the data')
        return True
    else:
        return False



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
import logging
import pydarn
import os

pydarn_log = logging.getLogger('pydarn')


class HardwareFileNotFoundError(Exception):
    """
    This error is raised when a hardware file for the provided
    abbreviation is not found.
    """
    def __init__(self, abbrev):
        self.abbreviation = abbrev
        self.path = os.path.dirname(pydarn.__file__)
        self.message = "Hardware file for {abv} radar was not found"\
            " in {path}/utils/hdw/."\
            " Please ensure the abbreviation is correct and there"\
            " exists a hardware file for it by checking:"\
            " https://github.com/superdarn/hdw. If this error occurs when"\
            " installing pydarn, please make an issue on the github page:"\
            "https://github.com/SuperDARN/pydarn/issues/new so that"\
            " developers are aware and can fix the problem."\
            " Please note hardware files are not obtained by RST."\
            "".format(abv=self.abbreviation, path=self.path)
        super().__init__(self.message)
        pydarn_log.error(self.message)


class RangeEstimationError(Exception):
    """
    This error is raised when there is an incorrect range estimation
    used.
    """
    def __init__(self, msg):
        self.message = msg
        super().__init__(self.message)
        pydarn_log.error(self.message)

